# -*- coding: utf-8 -*-

"""Pre-process visibilities into a form that is more suitable for gridding.

The following transformations are done:

- Autocorrelations are removed
- Visibilities with w < 0 are flipped
- UVW coordinates are converted to grid coordinates and quantized
- Weights are multiplied into the visibilities (but are still retained)
- Visibilities are sorted by channel then W slice
- Visibilities are partially sorted by baseline, then time
- Visibilities with the same UVW coordinates (after quantisation) are merged
  ("compressed").

The resulting visibilities are stored in HDF5
(:class:`VisibilityCollectorHDF5`) or numpy arrays
(:class:`VisibilityCollectorMem`). The latter is used mainly for testing.

Partial sorting by baseline is implemented by buffering up a reasonably large
number of visibilities and sorting them. The merging is also partial, since
only adjacent visibilities are candidates for merging.

The core pieces are implemented in C++ for speed (see preprocess.cpp).
"""

import sys
import math
import logging

import h5py
import numpy as np
from astropy import units

from katsdpimager import _preprocess


logger = logging.getLogger(__name__)


def _make_dtype(num_polarizations):
    """Creates a numpy structured dtype to hold a preprocessed visibility with
    associated metadata.

    Parameters
    ----------
    num_polarizations : int
        Number of polarizations in the visibility.
    """
    fields = [
        ('uv', np.int16, (2,)),
        ('sub_uv', np.int16, (2,)),
        ('weights', np.float32, (num_polarizations,)),
        ('vis', np.complex64, (num_polarizations,)),
        ('w_plane', np.int16)
    ]
    return np.dtype(fields)


def _make_fapl(cache_entries, cache_size, w0):
    """Create a File Access Properties List for h5py with a specified number
    of cache entries and cache size. This is based around the internal
    make_fapl function in h5py.
    """

    fapl = h5py.h5p.create(h5py.h5p.FILE_ACCESS)
    cache_settings = list(fapl.get_cache())
    fapl.set_cache(cache_settings[0], cache_entries, cache_size, w0)
    fapl.set_fclose_degree(h5py.h5f.CLOSE_STRONG)
    fapl.set_libver_bounds(h5py.h5f.LIBVER_LATEST, h5py.h5f.LIBVER_LATEST)
    return fapl


class VisibilityCollector(_preprocess.VisibilityCollector):
    """Base class that accepts a stream of visibility data and stores it. The
    subclasses provide the storage backends. Multiple channels are supported.

    Parameters
    ----------
    image_parameters : list of :class:`katsdpimager.parameters.ImageParameters`
        The image parameters for each channel. They must all have the same set
        of polarizations.
    grid_parameters : list of :class:`katsdpimager.parameters.GridParameters`
        Gridding parameters, with one entry per entry in `image_parameters`
    buffer_size : int
        Number of visibilities to buffer, prior to compression
    """
    def __init__(self, image_parameters, grid_parameters, buffer_size):
        num_polarizations = len(image_parameters[0].polarizations)
        if len(image_parameters) != len(grid_parameters):
            raise ValueError('Inconsistent lengths of image_parameters and grid_parameters')
        num_channels = len(image_parameters)
        config = np.zeros(num_channels, _preprocess.CHANNEL_CONFIG_DTYPE)
        for i, grid_p in enumerate(grid_parameters):
            config[i]["max_w"] = grid_p.max_w.to(units.m).value
            config[i]["w_slices"] = grid_p.w_slices
            config[i]["w_planes"] = grid_p.w_planes
            config[i]["oversample"] = grid_p.oversample
            config[i]["cell_size"] = image_parameters[i].cell_size.to(units.m).value
        super().__init__(
            num_polarizations, config, self._emit, buffer_size)
        self.image_parameters = image_parameters
        self.grid_parameters = grid_parameters
        self.store_dtype = _make_dtype(num_polarizations)

    @property
    def num_channels(self):
        return len(self.image_parameters)

    def _emit(self, elements):
        """Write an array of compressed elements with the same channel and w
        slice to the backing store. The caller must provide a non-empty
        array.

        Overrides of this function *must not* hold a reference to `elements`.
        The memory underneath is not reference-counted and is invalid once
        this function returns.
        """
        raise NotImplementedError()

    def add(self, uvw, weights, baselines, vis,
            feed_angle1, feed_angle2, mueller_stokes, mueller_circular):
        """Add a set of visibilities to the collector. Each of the provided
        arrays must have the same size on the N axis.

        Parameters
        ----------
        uvw : Quantity array, float32
            N×3 array of UVW coordinates.
        weights : array, float32
            C×N×Q array of weights, where C is the number of channels and
            Q is the number of input polarizations. Flags must be folded into
            the weights.
        baselines : array, int
            1D array of integer, indicating baseline IDs. The IDs are
            arbitrary and need not be contiguous, and are used only to
            associate visibilities from the same baseline together.
            Negative baseline IDs indicate autocorrelations, which will
            be discarded.
        vis : array, complex64
            C×N×Q array of visibilities.
        feed_angle1, feed_angle2 : array, float32
            Feed angles in radians for the two antennas in the baseline,
            for rotating from feed-relative to celestial frame.
        mueller_stokes : matrix
            Converts from circular polarization to output Stokes parameters (4×Q)
        mueller_circular : matrix
            Converts from input polarizations to circular frame (P×4, where P
            is the number of output polarizations)
        """

        # Ensure metres
        uvw = units.Quantity(uvw, unit=units.m, copy=False)
        uvw = np.require(uvw.value, np.float32, 'C')
        super().add(
            uvw, weights, baselines, vis, feed_angle1, feed_angle2,
            mueller_stokes, mueller_circular)

    def reader(self):
        """Create and return a reader object that can be used to iterate over
        visibility data. This may only be called *after* :meth:`close`.
        """
        raise NotImplementedError()


def _is_prime(n):
    for i in range(2, int(math.sqrt(n) + 1)):
        if n % i == 0:
            return False
    return True


class VisibilityCollectorHDF5(VisibilityCollector):
    """Visibility collector that stores data in an HDF5 file. All the
    visibilities are stored in one dataset, with axes for channel and W-slice.
    The visibilities for one channel and W-slice are strung along the third
    axis. This creates a ragged array, which HDF5 does not directly support
    (it supports variable-length arrays, but as far as I can tell each VL
    array is stored and retrieved as a unit, rather than chunked). However,
    chunked storage allocates chunks only as needed, so we store a regular
    array and use an attribute to indicate the length of each row.

    An alternative is to use a separate dataset for each channel and W-slice.
    This will work, but (again, as far as I can tell), libhdf5 uses a separate
    cache for each dataset, making it difficult to bound the total memory
    usage in this scenario.

    Chunks are shaped so that each chunk only contains data for one channel
    and W-slice, to allow this data to be efficiently read back.

    Parameters
    ----------
    filename : str
        Filename for HDF5 file to write
    max_cache_size : int, optional
        Maximum bytes to use for the HDF5 chunk cache. The default is
        unbounded. The actual size will not be larger than one chunk per
        channel/w-slice pair.
    chunk_elements : int, optional
        Number of visibilities per chunk. The default corresponds to about
        1MB per chunk for full-Stokes. Reducing this will save memory, but may
        reduce performance (particularly for spinning drives with high
        latency).
    args,kwargs
        Passed to base class constructor
    """

    def __init__(self, filename, *args, **kwargs):
        max_cache_size = kwargs.pop('max_cache_size', None)
        chunk_elements = kwargs.pop('chunk_elements', 16384)
        super().__init__(*args, **kwargs)
        chunk_size = chunk_elements * self.store_dtype.itemsize
        # We will be jumping between channels and W slices, so to avoid
        # evicting a chunk and then reloading it, we ideally have a big
        # enough cache to hold one chunk for each channel+w-slice.
        cache_chunks = sum(grid_p.w_slices for grid_p in self.grid_parameters)
        max_w_slices = max(grid_p.w_slices for grid_p in self.grid_parameters)
        cache_size = chunk_size * cache_chunks
        if max_cache_size is not None and cache_size > max_cache_size:
            cache_size = max_cache_size
            cache_chunks = cache_size // chunk_size
        # HDF5 recommendation for max performance is 100 slots per chunk, and
        # a prime number
        slots = cache_chunks * 100 + 1
        while not _is_prime(slots):
            slots += 2
        logger.debug('Setting cache size to %d slots, %d bytes', slots, cache_size)
        if isinstance(filename, str):
            filename = filename.encode(sys.getfilesystemencoding())
        self.filename = filename
        self._file = h5py.File(h5py.h5f.create(filename, fapl=_make_fapl(slots, cache_size, 1.0)))
        self._length = np.zeros((self.num_channels, max_w_slices), np.int64)
        self._dataset = self._file.create_dataset(
            "vis", (self.num_channels, max_w_slices, 0),
            maxshape=(self.num_channels, max_w_slices, None),
            dtype=self.store_dtype,
            compression='gzip',
            chunks=(1, 1, chunk_elements))

    def _emit(self, elements):
        N = elements.shape[0]
        channel = elements[0]['channel']
        w_slice = elements[0]['w_slice']
        old_length = self._length[channel, w_slice]
        self._length[channel, w_slice] += N
        if self._length[channel, w_slice] > self._dataset.shape[2]:
            self._dataset.resize(self._length[channel, w_slice], axis=2)
        # Work around FutureWarning in numpy 1.12 by first creating a view of
        # elements that contains only the fields we want.
        elements = np.ndarray(elements.shape, self.store_dtype, elements, 0, elements.strides)
        # This slightly contorted access is for performance reasons: see
        # https://github.com/h5py/h5py/issues/492
        self._dataset[channel : channel+1,
                      w_slice : w_slice+1,
                      old_length : self._length[channel, w_slice]] = \
            elements.astype(self.store_dtype)[np.newaxis, np.newaxis, :]

    def close(self):
        super().close()
        self._dataset.attrs.create('length', self._length)
        if logger.isEnabledFor(logging.INFO):
            filesize = self._file.id.get_filesize()
            n_compressed = np.sum(self._length)
            expected = self.store_dtype.itemsize * n_compressed
            if expected > 0:
                logger.info("Wrote %d visibilities in %d bytes to %s (%.2f%% compression ratio)",
                            n_compressed, filesize, self._file.filename,
                            100.0 * filesize / expected)
            else:
                logger.info("Wrote %d bytes to %s (no visibilities)",
                            filesize, self._file.filename)
        self._file.close()
        self._file = None

    def reader(self):
        return VisibilityReaderHDF5(self)


class VisibilityCollectorMem(VisibilityCollector):
    """Visibility collector that stores data in memory. Each dataset is stored
    as a list of numpy arrays.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.datasets = [
            [[] for w_slice in range(self.grid_parameters[channel].w_slices)]
            for channel in range(self.num_channels)]

    def _emit(self, elements):
        dataset = self.datasets[elements[0]['channel']][elements[0]['w_slice']]
        # Work around FutureWarning in numpy 1.12 by first creating a view of
        # elements that contains only the fields we want.
        elements = np.ndarray(elements.shape, self.store_dtype, elements, 0, elements.strides)
        dataset.append(elements.astype(self.store_dtype))

    def reader(self):
        return VisibilityReaderMem(self)


class VisibilityReader:
    """Abstract base class for visibility readers.

    Parameters
    ----------
    collector : subclass of :class:`VisibilityCollector`
        Closed collector that collected the visibilities.
    """
    def __init__(self, collector):
        pass

    def iter_slice(self, channel, w_slice, block_size=None):
        """A generator that iterates over the visibilities in blocks. Each
        iteration yields an array of some number of visibilities, in the type
        returned by :func:`_make_dtype`.

        If `block_size` is specified, the visibilities will be batched into
        groups of this size (except possibly the final batch, which will be
        shorter). If not specified, batches are of arbitrary length,
        depending on the collector.

        .. warning:: An implementation may recycle a single buffer for this
           purpose. The returned array must be consumed (or copied) before
           requesting the next one.
        """
        raise NotImplementedError()

    def len(self, channel, w_slice):
        """Number of visibilities that will be enumerated by :meth:`iter_slice`"""
        raise NotImplementedError()

    def close(self):
        """Free any resources associated with the reader."""
        pass


class VisibilityReaderHDF5(VisibilityReader):
    def __init__(self, collector):
        super().__init__(collector)
        # We're doing a linear read over the file, so we don't need to increase
        # the cache size.
        # TODO: experiment with setting a tiny cache so that blocks bypass the
        # cache entirely.
        self._file = h5py.File(h5py.h5f.open(collector.filename, flags=h5py.h5f.ACC_RDONLY))
        self._dataset = self._file["vis"]
        self._length = self._dataset.attrs['length']

    def len(self, channel, w_slice):
        return self._length[channel, w_slice]

    def iter_slice(self, channel, w_slice, block_size=None):
        if block_size is None:
            block_size = self._dataset.chunks[2]
        buf3d = np.rec.recarray((1, 1, block_size), self._dataset.dtype)
        buf = buf3d[0, 0, :]
        N = self.len(channel, w_slice)
        for start in range(0, N - block_size + 1, block_size):
            self._dataset.read_direct(
                buf3d,
                np.s_[channel : channel+1, w_slice : w_slice+1, start : start+block_size],
                np.s_[:, :, 0 : block_size])
            yield buf
        last = N % block_size
        if last > 0:
            self._dataset.read_direct(
                buf3d,
                np.s_[channel : channel+1, w_slice : w_slice+1, N - last : N],
                np.s_[:, :, 0 : last])
            yield buf[:last]

    @property
    def num_channels(self):
        return self._length.shape[0]

    def num_w_slices(self, channel):
        return self._length.shape[1]

    def close(self):
        super().close()
        self._file.close()
        self._file = None


class VisibilityReaderMem(VisibilityReader):
    def __init__(self, collector):
        super().__init__(collector)
        self.datasets = collector.datasets

    def _iter_slice_blocked(self, channel, w_slice, block_size):
        dataset = self.datasets[channel][w_slice]
        if not dataset:
            return
        buf = np.rec.recarray((block_size,), dataset[0].dtype)
        buf_pos = 0
        for dset in dataset:
            dset_pos = 0
            while len(dset) - dset_pos > block_size - buf_pos:
                buf[buf_pos:] = dset[dset_pos : dset_pos + (block_size - buf_pos)]
                yield buf
                dset_pos += block_size - buf_pos
                buf_pos = 0
            buf[buf_pos : buf_pos + len(dset) - dset_pos] = dset[dset_pos:]
            buf_pos += len(dset) - dset_pos
        if buf_pos > 0:
            yield buf[:buf_pos]

    def iter_slice(self, channel, w_slice, block_size=None):
        if block_size is None:
            return iter(self.datasets[channel][w_slice])
        else:
            return self._iter_slice_blocked(channel, w_slice, block_size)

    def len(self, channel, w_slice):
        return sum(len(x) for x in self.datasets[channel][w_slice])

    @property
    def num_channels(self):
        return len(self.datasets)

    def num_w_slices(self, channel):
        return len(self.datasets[channel])

    def close(self):
        super().close()
        self.datasets = None



"""Deconvolution routines based on CLEAN.

Both the CPU and GPU versions use an acceleration structure to speed up
peak-finding when the point spread function (PSF) patch in use is
significantly smaller than the image. The image is divided into tiles. For
each tile, the peak value and the position of that value are maintained.
Finding the global peak requires only finding the best tile peak. Subtracting
the PSF requires only updating those tiles intersected by the shifted PSF.

The GPU implementation currently round-trips to the CPU on each minor cycle.
It could be done entirely on the GPU, but round-tripping will make it easier
to put in a threshold later. It should still be possible to do batches of
minor cycles if the launch overheads become an issue.

.. include:: macros.rst
"""

import math

import numpy as np
from katsdpsigproc import accel
import pkg_resources

import katsdpimager.types
from katsdpimager import numba

#: Use only Stokes I to find peaks
CLEAN_I = 0
#: Use the sum of squares of available Stokes components to find peaks
CLEAN_SUMSQ = 1


class PsfPatchTemplate:
    """Examines a PSF to determine how big a PSF patch is required to
    enclosure all values above some threshold.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : {`np.float32`, `np.float64`}
        Precision of image
    num_polarizations : int
        Number of polarizations stored in the dirty image
    tuning : dict, optional
        Tuning parameters (unused for now)
    """
    def __init__(self, context, dtype, num_polarizations, tuning=None):
        # TODO: autotuning
        self.num_polarizations = num_polarizations
        self.dtype = dtype
        self.wgsx = 16
        self.wgsy = 16
        self.program = accel.build(
            context, "imager_kernels/clean/psf_patch.mako",
            {
                'real_type': katsdpimager.types.dtype_to_ctype(dtype),
                'wgsx': self.wgsx,
                'wgsy': self.wgsy,
                'num_polarizations': num_polarizations
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return PsfPatch(self, *args, **kwargs)


class PsfPatch(accel.Operation):
    """Instantiation of :class:`PsfPatchTemplate`.

    The implementation works by dividing the PSF into tiles, with a workgroup
    per tile, and doing a workgroup-level reduction in each tile. The
    inter-tile reduction is done on the host.

    The returned patch size is generally odd, but will never exceed the PSF size
    (so if the side lobes extend all the way to the edge, it will be equal to the
    PSF size, which is even).

    .. rubric:: Slots

    **psf** : array of shape(num_polarizations, width, height), real
        PSF, already scaled so that the central value is 1.
    **bound** : array, int
        Array of implementation-dependent size containing lower bounds on the
        l and m distance from the centre, in pixels.

    Parameters
    ----------
    template : :class:`PsfPatchTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple of ints
        Shape for the PSF
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """

    def __init__(self, template, command_queue, shape, allocator=None):
        if shape[0] != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        super().__init__(command_queue, allocator)
        max_wg_x = accel.divup(shape[2], template.wgsx)
        max_wg_y = accel.divup(shape[1], template.wgsy)
        polarizations = accel.Dimension(template.num_polarizations, exact=True)
        self.slots['psf'] = accel.IOSlot([polarizations, shape[1], shape[2]], template.dtype)
        self.slots['bound'] = accel.IOSlot([
            accel.Dimension(max_wg_y * max_wg_x),
            accel.Dimension(2, exact=True)], np.int32)
        self._bound_host = accel.HostArray((max_wg_y * max_wg_x, 2), np.int32,
                                           context=command_queue.context)
        self.template = template
        self.kernel = template.program.get_kernel('psf_patch')

    def _run(self):
        # Never used because we override __call__
        pass        # pragma: nocover

    def __call__(self, threshold, limit=None, **kwargs):
        self.bind(**kwargs)
        self.ensure_all_bound()
        psf = self.buffer('psf')
        bound = self.buffer('bound')
        min_x = 0
        min_y = 0
        max_x = psf.shape[2] - 1
        max_y = psf.shape[1] - 1
        mid_x = psf.shape[2] // 2
        mid_y = psf.shape[1] // 2
        if limit is not None:
            hlimit = (limit - 1) // 2
            min_x = max(min_x, mid_x - hlimit)
            min_y = max(min_y, mid_y - hlimit)
            max_x = min(max_x, mid_x + hlimit)
            max_y = min(max_y, mid_y + hlimit)
        wg_x = accel.divup(max_x - min_x + 1, self.template.wgsx)
        wg_y = accel.divup(max_y - min_y + 1, self.template.wgsy)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                psf.buffer,
                np.int32(psf.padded_shape[2]),
                np.int32(psf.padded_shape[1] * psf.padded_shape[2]),
                bound.buffer,
                np.int32(min_x), np.int32(min_y),
                np.int32(max_x), np.int32(max_y),
                np.int32(mid_x), np.int32(mid_y),
                np.float32(threshold)
            ],
            global_size=(wg_x * self.template.wgsx, wg_y * self.template.wgsy),
            local_size=(self.template.wgsx, self.template.wgsy)
        )
        if isinstance(bound, accel.SVMArray):
            self.command_queue.finish()
        bound.get(self.command_queue, self._bound_host)
        # Turn distances from the centre into a symmetric bounding box size.
        box = 2 * np.max(self._bound_host[: wg_x * wg_y], axis=0) + 1
        return (self.template.num_polarizations,
                int(min(box[1], psf.shape[1])),
                int(min(box[0], psf.shape[2])))


def metric_to_power(mode, metric):
    """Convert a peak-finding metric to a value that scales linearly with
    power (e.g. Jy/beam).
    """
    if mode == CLEAN_I:
        return metric
    elif mode == CLEAN_SUMSQ:
        return math.sqrt(metric)
    else:
        raise ValueError('Invalid mode {}'.format(mode))


def power_to_metric(mode, power):
    """Inverse of :func:`metric_to_power`."""
    if mode == CLEAN_I:
        return power
    elif mode == CLEAN_SUMSQ:
        return power * power
    else:
        raise ValueError('Invalid mode {}'.format(mode))


class NoiseEstTemplate:
    """Robust estimation of the noise (as a standard deviation) in an image.

    The noise is estimated by computing the median absolute value via binary
    search.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : {`np.float32`, `np.float64`}
        Image precision
    num_polarizations : int
        number of polarizations stored in the dirty/residual image
    mode : {:data:`CLEAN_I`, :data:`CLEAN_SUMSQ`}
        Metric for scoring pixels
    tuning : dict, optional
        Tuning parameters (unused for now)
    """
    def __init__(self, context, dtype, num_polarizations, mode, tuning=None):
        self.context = context
        self.dtype = np.dtype(dtype)
        self.num_polarizations = num_polarizations
        self.mode = mode
        self.wgsx = 32
        self.wgsy = 8
        self.tilex = 32
        self.tiley = 32
        self.program = accel.build(
            context, "imager_kernels/clean/rank.mako",
            {
                'real_type': katsdpimager.types.dtype_to_ctype(dtype),
                'wgsx': self.wgsx,
                'wgsy': self.wgsy,
                'tilex': self.tilex,
                'tiley': self.tiley,
                'num_polarizations': num_polarizations,
                'clean_mode': mode
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return NoiseEst(self, *args, **kwargs)


class NoiseEst(accel.Operation):
    """Instantiation of :class:`NoiseEstTemplate`

    .. rubric:: Slots

    **dirty** : array of shape (num_polarizations, height, width), real
        Dirty image
    **rank** : implementation-defined
        Temporary buffer used during operation. It holds a rank of a test
        value for each workgroup.

    Parameters
    ----------
    template : :class:`NoiseEstTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    image_shape : tuple of ints
        Shape for the dirty image
    border : int
        Distance from each edge of dirty image to ignore in ranking
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """

    def __init__(self, template, command_queue, image_shape, border, allocator=None):
        if image_shape[0] != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        if border * 2 >= min(image_shape[1], image_shape[2]):
            raise ValueError('Border must be less than half the image size')
        super().__init__(command_queue, allocator)
        self._num_tiles_x = accel.divup(image_shape[2] - 2 * border, template.tilex)
        self._num_tiles_y = accel.divup(image_shape[1] - 2 * border, template.tiley)
        self.template = template
        self.border = border
        self.slots['dirty'] = accel.IOSlot([
            accel.Dimension(template.num_polarizations, exact=True),
            image_shape[1], image_shape[2]], template.dtype)
        self.slots['rank'] = accel.IOSlot([
            accel.Dimension(self._num_tiles_y, exact=True),
            accel.Dimension(self._num_tiles_x, exact=True)], np.uint32)
        self.kernel = template.program.get_kernel('compute_rank')

    def _run(self):
        # Never used because we override __call__
        pass        # pragma: nocover

    def __call__(self, **kwargs):
        self.bind(**kwargs)
        self.ensure_all_bound()
        dtype = self.template.dtype
        if dtype == np.float32:
            itype = np.uint32
        elif dtype == np.float64:
            itype = np.uint64
        else:
            raise TypeError('dtype {} is not supported'.format(dtype))
        # We do binary search on the bit representation of the floating point
        # value.
        low = dtype.type(0)
        high = dtype.type(np.inf)
        dirty = self.buffer('dirty')
        rank = self.buffer('rank')
        rank_host = rank.empty_like()
        median_rank = (dirty.shape[1] - 2 * self.border) * (dirty.shape[2] - 2 * self.border) // 2
        # We don't need a super-accurate estimate down to the last bit.
        while high > np.finfo(dtype).tiny and high > low * 1.0001:
            # Average the integer bit representations, which will effectively
            # binary search the exponent.
            ilow = low.view(itype)
            ihigh = high.view(itype)
            if ihigh - ilow == itype(1):
                # This can happen in some corner cases despite the ratio chunk
                # e.g. if low is zero or high is inf
                break
            imid = ilow + (ihigh - ilow) // itype(2)
            mid = imid.view(dtype)
            self.command_queue.enqueue_kernel(
                self.kernel,
                [
                    dirty.buffer,
                    np.int32(dirty.padded_shape[2]),
                    np.int32(dirty.padded_shape[1] * dirty.padded_shape[2]),
                    np.int32(dirty.shape[2] - self.border),
                    np.int32(dirty.shape[1] - self.border),
                    np.int32(self.border),
                    rank.buffer,
                    mid
                ],
                global_size=(self._num_tiles_x * self.template.wgsx,
                             self._num_tiles_y * self.template.wgsy),
                local_size=(self.template.wgsx, self.template.wgsy))
            if isinstance(rank, accel.SVMArray):
                self.command_queue.finish()
            rank.get(self.command_queue, rank_host)
            cur_rank = rank_host.sum()
            if cur_rank < median_rank:
                low = mid
            else:
                high = mid
        # low and high are very close, but we use low so that if the input is
        # more than 50% zeros then the output is zero.
        # The magic number is the ratio between median absolute deviation and
        # standard deviation of a Gaussian.
        return metric_to_power(self.template.mode, low) * 1.48260222


class _UpdateTilesTemplate:
    """Operation template to compute the peak (including location) for each
    tile intersecting a window.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : {`np.float32`, `np.float64`}
        Precision of image
    num_polarizations : int
        Number of polarizations stored in the dirty image
    mode : {:data:`CLEAN_I`, :data:`CLEAN_SUMSQ`}
        Function determining peak
    tuning : dict, optional
        Tuning parameters (unused for now)
    """
    def __init__(self, context, dtype, num_polarizations, mode, tuning=None):
        # TODO: autotuning
        self.num_polarizations = num_polarizations
        self.dtype = dtype
        self.wgsx = 32
        self.wgsy = 8
        self.tilex = 32
        self.tiley = 32
        self.program = accel.build(
            context, "imager_kernels/clean/update_tiles.mako",
            {
                'real_type': katsdpimager.types.dtype_to_ctype(dtype),
                'wgsx': self.wgsx,
                'wgsy': self.wgsy,
                'tilex': self.tilex,
                'tiley': self.tiley,
                'num_polarizations': num_polarizations,
                'clean_mode': mode
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return _UpdateTiles(self, *args, **kwargs)


class _UpdateTiles(accel.Operation):
    """Instantiation of :class:`_UpdateTileTemplate`

    .. rubric:: Slots

    **dirty** : array of shape (num_polarizations, height, width), real
        Dirty image
    **tile_max** : array of shape (height, width)
        Internal storage of maximum score for each tile
    **tile_pos** : array of shape (height, width, 2)
        Internal storage of best position for each tile, with each position
        stored as (row, col).

    Parameters
    ----------
    template : :class:`_UpdateTilesTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    image_shape : tuple of ints
        Shape for the dirty image
    border : int
        Distance from each edge of dirty image where tiles start
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """

    def __init__(self, template, command_queue, image_shape, border, allocator=None):
        if image_shape[0] != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        if border * 2 >= min(image_shape[1], image_shape[2]):
            raise ValueError('Border must be less than half the image size')
        super().__init__(command_queue, allocator)
        num_tiles_x = accel.divup(image_shape[2] - 2 * border, template.tilex)
        num_tiles_y = accel.divup(image_shape[1] - 2 * border, template.tiley)
        image_width = accel.Dimension(image_shape[2])
        image_height = accel.Dimension(image_shape[1])
        image_pols = accel.Dimension(template.num_polarizations, exact=True)
        tiles_width = accel.Dimension(num_tiles_x)
        tiles_height = accel.Dimension(num_tiles_y)
        self.template = template
        self.border = border
        self.slots['dirty'] = accel.IOSlot([image_pols, image_height, image_width], template.dtype)
        self.slots['tile_max'] = accel.IOSlot([tiles_height, tiles_width], template.dtype)
        self.slots['tile_pos'] = accel.IOSlot(
            [tiles_height, tiles_width, accel.Dimension(2, exact=True)], np.int32)
        self.kernel = template.program.get_kernel('update_tiles')

    def _run(self):
        # Never used because we override __call__
        pass        # pragma: nocover

    def __call__(self, x0, y0, x1, y1, **kwargs):
        """Update all tiles intersected by the pixel range [x0, x1) by [y0, y1)"""
        self.bind(**kwargs)
        self.ensure_all_bound()
        tile_max = self.buffer('tile_max')
        tile_pos = self.buffer('tile_pos')
        # Map pixel range to a tile range
        x0 = max((x0 - self.border) // self.template.tilex, 0)
        y0 = max((y0 - self.border) // self.template.tiley, 0)
        x1 = min(accel.divup(x1 - self.border, self.template.tilex), tile_max.shape[1])
        y1 = min(accel.divup(y1 - self.border, self.template.tiley), tile_max.shape[0])
        if x0 < x1 and y0 < y1:
            dirty = self.buffer('dirty')
            self.command_queue.enqueue_kernel(
                self.kernel,
                [
                    dirty.buffer,
                    np.int32(dirty.padded_shape[2]),
                    np.int32(dirty.padded_shape[1] * dirty.padded_shape[2]),
                    np.int32(dirty.shape[2] - self.border),
                    np.int32(dirty.shape[1] - self.border),
                    np.int32(self.border),
                    tile_max.buffer,
                    tile_pos.buffer,
                    np.int32(tile_max.padded_shape[1]),
                    np.int32(x0), np.int32(y0)
                ],
                global_size=((x1 - x0) * self.template.wgsx, (y1 - y0) * self.template.wgsy),
                local_size=(self.template.wgsx, self.template.wgsy))


class _FindPeakTemplate:
    """Find the global peak from per-tile peaks.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : {`np.float32`, `np.float64`}
        Image precision
    tuning : dict, optional
        Tuning parameters (currently unused)
    """

    def __init__(self, context, dtype, num_polarizations, tuning=None):
        # TODO: autotuning
        self.wgsx = 16
        self.wgsy = 16
        self.dtype = dtype
        self.num_polarizations = num_polarizations
        self.program = accel.build(
            context, "imager_kernels/clean/find_peak.mako",
            {
                'real_type': katsdpimager.types.dtype_to_ctype(dtype),
                'num_polarizations': self.num_polarizations,
                'wgsx': self.wgsx,
                'wgsy': self.wgsy,
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return _FindPeak(self, *args, **kwargs)


class _FindPeak(accel.Operation):
    """Instantiation of :class:`_FindPeak`.

    .. rubric:: Slots

    **dirty** : array of shape (num_polarizations, height, width), real
        Dirty image, used to return the original values at the peak
    **tile_max** : array of shape (height, width)
        Internal storage of maximum score for each tile
    **tile_pos** : array of shape (height, width, 2)
        Internal storage of best position for each tile, with each position
        stored as (row, col).
    **peak_value** : array of shape (1,), real
        Score for the peak position (output)
    **peak_pos** : array of shape (2,), int32
        Row and column of the peak position (output)
    **peak_pixel** : array of shape (num_polarizations,), real
        Dirty image sampled at the peak (output)

    Parameters
    ----------
    template : :class:`_FindPeakTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    image_shape : tuple of int
        Shape of the dirty image, as (num_polarizations, height, width)
    tile_shape : tuple of int
        Shape of the tile array, as (height, width)
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, image_shape, tile_shape, allocator=None):
        if image_shape[0] != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        super().__init__(command_queue, allocator)
        self.template = template
        image_dims = [accel.Dimension(image_shape[0], exact=True),
                      accel.Dimension(image_shape[1]),
                      accel.Dimension(image_shape[2])]
        tile_dims = [accel.Dimension(tile_shape[0]), accel.Dimension(tile_shape[1])]
        pair = accel.Dimension(2, exact=True)
        self.slots['dirty'] = accel.IOSlot(image_dims, template.dtype)
        self.slots['tile_max'] = accel.IOSlot(tile_dims, template.dtype)
        self.slots['tile_pos'] = accel.IOSlot(tile_dims + [pair], np.int32)
        self.slots['peak_value'] = accel.IOSlot([1], template.dtype)
        self.slots['peak_pos'] = accel.IOSlot([2], np.int32)
        self.slots['peak_pixel'] = accel.IOSlot([template.num_polarizations], template.dtype)
        self.kernel = template.program.get_kernel('find_peak')

    def _run(self):
        dirty = self.buffer('dirty')
        tile_max = self.buffer('tile_max')
        tile_pos = self.buffer('tile_pos')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                dirty.buffer,
                np.int32(dirty.padded_shape[2]),
                np.int32(dirty.padded_shape[1] * dirty.padded_shape[2]),
                tile_max.buffer,
                tile_pos.buffer,
                np.int32(tile_max.padded_shape[1]),
                np.int32(tile_max.shape[1]), np.int32(tile_max.shape[0]),
                self.buffer('peak_value').buffer,
                self.buffer('peak_pos').buffer,
                self.buffer('peak_pixel').buffer
            ],
            global_size=(accel.roundup(tile_max.shape[0], self.template.wgsx),
                         accel.roundup(tile_max.shape[1], self.template.wgsy)),
            local_size=(self.template.wgsx, self.template.wgsy))


class _SubtractPsfTemplate:
    """Subtract a multiple of the point spread function from the dirty image,
    and add a corresponding single pixel to the model image.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : {`np.float32`, `np.float64`}
        Image precision
    num_polarizations : int
        number of polarizations stored in the dirty image and PSF
    tuning : dict, optional
        Tuning parameters (unused for now)
    """
    def __init__(self, context, dtype, num_polarizations, tuning=None):
        # TODO: autotuning
        self.wgsx = 16
        self.wgsy = 16
        self.dtype = dtype
        self.num_polarizations = num_polarizations
        self.program = accel.build(
            context, "imager_kernels/clean/subtract_psf.mako",
            {
                'real_type': katsdpimager.types.dtype_to_ctype(dtype),
                'num_polarizations': num_polarizations,
                'wgsx': self.wgsx,
                'wgsy': self.wgsy
            },
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return _SubtractPsf(self, *args, **kwargs)


class _SubtractPsf(accel.Operation):
    """Instantiation of :class:`_SubtractPsfTemplate`. The dirty and model
    image have the same shape and padding, while the PSF may be smaller.

    .. rubric:: Slots

    **dirty** : array of shape (num_polarizations, height, width), real
        Dirty image
    **model** : array of shape (num_polarizations, height, width), real
        Model image
    **psf** : array of shape (num_polarizations, height, width), real
        Point spread function, with center at (height // 2, width // 2)
    **peak_pixel** : array of shape (num_polarizations,), real
        Current dirty image value at the position where subtract is happening

    Parameters
    ----------
    template : :class:`_SubtractPsfTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    loop_gain : float
        Scale factor for subtraction. The PSF is scaled by both `loop_gain` and
        the per-polarization value in the **peak_pixel** slot before subtraction.
    image_shape : tuple of int
        Shape for the dirty and model images, as (num_polarizations, height, width)
    psf_shape : tuple of int
        Shape for the point spread function, as (num_polarizations, height, width)
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if the number of polarizations in `image_shape` or `psf_shape` does not
        match `template`
    """
    def __init__(self, template, command_queue, loop_gain, image_shape, psf_shape, allocator=None):
        super().__init__(command_queue, allocator)
        pol_dim = accel.Dimension(template.num_polarizations, exact=True)
        if image_shape[0] != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        if psf_shape[0] != template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        image_dims = [pol_dim, accel.Dimension(image_shape[1]), accel.Dimension(image_shape[2])]
        psf_dims = [pol_dim, accel.Dimension(psf_shape[1]), accel.Dimension(psf_shape[2])]
        self.slots['dirty'] = accel.IOSlot(image_dims, template.dtype)
        self.slots['model'] = accel.IOSlot(image_dims, template.dtype)
        self.slots['psf'] = accel.IOSlot(psf_dims, template.dtype)
        self.slots['peak_pixel'] = accel.IOSlot([pol_dim], template.dtype)
        self.loop_gain = loop_gain
        self.template = template
        self.kernel = template.program.get_kernel('subtract_psf')

    def _run(self):
        # Never used because we override __call__
        pass        # pragma: nocover

    def __call__(self, pos, psf_patch, **kwargs):
        """Execute the operation.

        Parameters
        ----------
        pos : tuple
            row, col in image at which to center the PSF
        psf_patch : tuple
            num_polarizations, height, width of central area of PSF to subtract
            (num_polarizations is ignored)
        """
        self.bind(**kwargs)
        self.ensure_all_bound()
        dirty = self.buffer('dirty')
        model = self.buffer('model')
        psf = self.buffer('psf')
        psf_y = psf.shape[1] // 2 - psf_patch[1] // 2
        psf_x = psf.shape[2] // 2 - psf_patch[2] // 2
        psf_addr_offset = psf_y * psf.padded_shape[2] + psf_x
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                dirty.buffer,
                model.buffer,
                np.int32(dirty.padded_shape[2]),
                np.int32(dirty.padded_shape[1] * dirty.padded_shape[2]),
                np.int32(dirty.shape[2]),
                np.int32(dirty.shape[1]),
                psf.buffer,
                np.int32(psf.padded_shape[2]),
                np.int32(psf.padded_shape[1] * psf.padded_shape[2]),
                np.int32(psf_patch[2]),
                np.int32(psf_patch[1]),
                np.int32(psf_addr_offset),
                self.buffer('peak_pixel').buffer,
                np.int32(pos[1]), np.int32(pos[0]),
                np.int32(pos[1] - psf_patch[2] // 2),
                np.int32(pos[0] - psf_patch[1] // 2),
                np.float32(self.loop_gain)
            ],
            global_size=(accel.roundup(psf_patch[2], self.template.wgsx),
                         accel.roundup(psf_patch[1], self.template.wgsy)),
            local_size=(self.template.wgsx, self.template.wgsy))


class CleanTemplate:
    """Composite template for the CLEAN minor cycles.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    clean_parameters : :class:`katsdpimager.parameters.CleanParameters`
        Command-line parameters for CLEAN
    dtype : {`np.float32`, `np.float64`}
        Image precision
    num_polarizations : int
        Number of polarizations stored in the image
    """
    def __init__(self, context, clean_parameters, dtype, num_polarizations):
        self.clean_parameters = clean_parameters
        self.dtype = dtype
        self.num_polarizations = num_polarizations
        self._update_tiles = _UpdateTilesTemplate(context, dtype, num_polarizations,
                                                  clean_parameters.mode)
        self._find_peak = _FindPeakTemplate(context, dtype, num_polarizations)
        self._subtract_psf = _SubtractPsfTemplate(context, dtype, num_polarizations)

    def instantiate(self, *args, **kwargs):
        return Clean(self, *args, **kwargs)


class Clean(accel.OperationSequence):
    """Instantiation of :class:`CleanTemplate`.

    .. rubric:: Slots

    **dirty** : array of shape (num_polarizations, height, width), real
        Dirty image
    **model** : array of shape (num_polarizations, height, width), real
        Model image
    **psf** : array of shape (num_polarizations, height, width), real
        Point spread function, with center at (height // 2, width // 2)
    **tile_max** : array of shape (height, width)
        Internal storage of maximum score for each tile
    **tile_pos** : array of shape (height, width, 2)
        Internal storage of best position for each tile, with each position
        stored as (row, col).
    **peak_value** : array of shape (1,), real
        Score for the peak position (output)
    **peak_pos** : array of shape (2,), int32
        Row and column of the peak position (output)
    **peak_pixel** : array of shape (num_polarizations,), real
        Dirty image sampled at the peak (output)

    The external interface consists of **dirty**, **model** and **psf**. The
    others are exposed only so that the memory can be reused for other
    purposes.

    Parameters
    ----------
    template : :class:`_UpdateTilesTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    image_parameters : :class:`katsdpimager.parameters.ImageParameters`
        Command-line parameters with image properties
    border : int
        Size of border in which no components may be placed
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if `image_parameters` is inconsistent with the template
    """
    def __init__(self, template, command_queue, image_parameters, allocator=None):
        if image_parameters.real_dtype != template.dtype:
            raise ValueError('dtype mismatch')
        image_shape = (len(image_parameters.polarizations),
                       image_parameters.pixels, image_parameters.pixels)
        self._update_tiles = template._update_tiles.instantiate(
            command_queue, image_shape, template.clean_parameters.border, allocator)
        tile_shape = self._update_tiles.slots['tile_max'].shape
        self._find_peak = template._find_peak.instantiate(
            command_queue, image_shape, tile_shape, allocator)
        self._subtract_psf = template._subtract_psf.instantiate(
            command_queue, template.clean_parameters.loop_gain, image_shape, image_shape,
            allocator)
        ops = [
            ('update_tiles', self._update_tiles),
            ('find_peak', self._find_peak),
            ('subtract_psf', self._subtract_psf)
        ]
        compounds = {
            'dirty': ['update_tiles:dirty', 'find_peak:dirty', 'subtract_psf:dirty'],
            'model': ['subtract_psf:model'],
            'psf': ['subtract_psf:psf'],
            'tile_max': ['update_tiles:tile_max', 'find_peak:tile_max'],
            'tile_pos': ['update_tiles:tile_pos', 'find_peak:tile_pos'],
            'peak_value': ['find_peak:peak_value'],
            'peak_pos': ['find_peak:peak_pos'],
            'peak_pixel': ['find_peak:peak_pixel', 'subtract_psf:peak_pixel']
        }
        super().__init__(command_queue, ops, compounds, allocator=allocator)
        peak_value = self.slots['peak_value']
        peak_pos = self.slots['peak_pos']
        self._peak_value_host = accel.HostArray(
            peak_value.shape, peak_value.dtype, peak_value.required_padded_shape(),
            context=command_queue.context)
        self._peak_pos_host = accel.HostArray(
            peak_pos.shape, peak_pos.dtype, peak_pos.required_padded_shape(),
            context=command_queue.context)

    def reset(self):
        """Call after populating the buffers but before the first minor cycle."""
        self.ensure_all_bound()
        dirty = self.buffer('dirty')
        self._update_tiles(0, 0, dirty.shape[2], dirty.shape[1])

    def __call__(self, psf_patch, threshold=0.0):
        """Run a single minor CLEAN cycle.

        Parameters
        ----------
        psf_patch : tuple
            Shape of the PSF patch to use for subtraction
        threshold : float, optional
            If specified, skip the cycle if the peak value metric is less
            than this threshold. Note that this value must be chosen relative
            to the clean mode, since the units are different.

        Returns
        -------
        peak_value : float
            The clean mode metric for the peak, or ``None`` if it was
            below the threshold and so no subtraction was done.
        """
        self.ensure_all_bound()
        self._find_peak()
        # Copy the peak position and value back to the host
        peak_value_device = self.buffer('peak_value')
        peak_pos_device = self.buffer('peak_pos')
        if (isinstance(peak_value_device, accel.SVMArray)
                or isinstance(peak_pos_device, accel.SVMArray)):
            self.command_queue.finish()
        peak_value = peak_value_device.get_async(self.command_queue, self._peak_value_host)
        peak_pos = peak_pos_device.get_async(self.command_queue, self._peak_pos_host)
        peak_pos = [int(x) for x in peak_pos]
        self.command_queue.finish()
        if peak_value[0] < threshold:
            return None

        self._subtract_psf(peak_pos, psf_patch)
        # Update the tiles
        x0 = peak_pos[1] - psf_patch[2] // 2
        x1 = x0 + psf_patch[2]
        y0 = peak_pos[0] - psf_patch[1] // 2
        y1 = y0 + psf_patch[1]
        self._update_tiles(x0, y0, x1, y1)
        return peak_value[0]


def psf_patch_host(psf, threshold, limit=None):
    """Compute the size of the bounding box of the PSF required to contain all
    values above `threshold`.

    This is a CPU-only equivalent to :class:`PsfPatch`.

    Parameters
    ----------
    psf : ndarray
        Point spread function, with shape (polarizations, height, width)
    threshold : float
        Minimum value that must be included inside the box
    limit : float, optional
        Upper bound on the return value. Pixels outside this range are not
        examined.

    Returns
    -------
    box : tuple
        Shape of the central part of `psf` to retain, with the same
        dimensions. The size is always even.
    """
    if limit is not None:
        hlimit = (limit - 1) // 2
        mid_x = psf.shape[2] // 2
        mid_y = psf.shape[1] // 2
        min_x = max(0, mid_x - hlimit)
        min_y = max(0, mid_y - hlimit)
        max_x = min(psf.shape[2] - 1, mid_x + hlimit)
        max_y = min(psf.shape[1] - 1, mid_y + hlimit)
        psf = psf[:, min_y : max_y + 1, min_x : max_x + 1]

    nz = np.nonzero(np.abs(psf) >= threshold)
    if len(nz[0]) == 0:
        # No values above threshold at all. This should never happen, because
        # the peak should be 1.0, but return a 1x1 box.
        return (psf.shape[0], 1, 1)
    y_dist = np.max(np.abs(nz[1] - psf.shape[1] // 2))
    x_dist = np.max(np.abs(nz[2] - psf.shape[2] // 2))
    y_size = min(psf.shape[1], 2 * y_dist + 1)
    x_size = min(psf.shape[2], 2 * x_dist + 1)
    return (psf.shape[0], y_size, x_size)


def noise_est_host(image, border, mode):
    """Host implementation of :class:`NoiseEstTemplate`."""
    image = image[:, border:-border, border:-border]
    if mode == CLEAN_I:
        metric = np.abs(image)
    elif mode == CLEAN_SUMSQ:
        metric = np.sum(image**2, axis=0)
    median = np.median(metric)
    return metric_to_power(mode, median) * 1.48260222


@numba.jit(nopython=True)
def _tile_peak(y0, x0, y1, x1, image, mode, zero):
    """Implementation of :meth:`CleanHost._update_tile`, split out as a
    function to allow numba to fully JIT it."""
    best_pos = (x0, y0)
    best_value = zero
    if mode == CLEAN_I:
        for y in range(y0, y1):
            for x in range(x0, x1):
                value = np.abs(image[0, y, x])
                if value > best_value:
                    best_value = value
                    best_pos = (y, x)
    else:
        for y in range(y0, y1):
            for x in range(x0, x1):
                value = zero
                for pol in range(image.shape[0]):
                    value += image[pol, y, x]**2
                if value > best_value:
                    best_value = value
                    best_pos = (y, x)
    return best_pos, best_value


class CleanHost:
    """CPU-only equivalent to :class:`Clean`. The class keeps references to
    the provided arrays, and only examines modifies them when :meth:`reset`
    or :meth:`__call__` is invoked.

    Parameters
    ----------
    image_parameters : :class:`katsdpimager.parameters.ImageParameters`
        Command-line parameters with image properties
    clean_parameters : :class:`katsdpimager.parameters.CleanParameters`
        Command-line parameters for CLEAN
    image : ndarray, (height, width, num_polarizations)
        Dirty image, modified as cleaning proceeds
    psf : ndarray, (height, width, num_polarizations), real
        Point spread function (may be smaller than `image`)
    model : ndarray, (height, width, num_polarizations), real
        Model image, which is updated with found CLEAN components
    """
    def __init__(self, image_parameters, clean_parameters, image, psf, model):
        self.clean_parameters = clean_parameters
        self.image_parameters = image_parameters
        self.image = image
        self.model = model
        self.psf = psf
        self.tile_size = 32
        border = self.clean_parameters.border
        tiles_x = accel.divup(image.shape[2] - 2 * border, self.tile_size)
        tiles_y = accel.divup(image.shape[1] - 2 * border, self.tile_size)
        self._tile_max = np.zeros((tiles_y, tiles_x), self.image_parameters.real_dtype)
        self._tile_pos = np.empty((tiles_y, tiles_x, 2), np.int32)

    def _update_tile(self, y, x):
        border = self.clean_parameters.border
        x0 = x * self.tile_size + border
        y0 = y * self.tile_size + border
        x1 = min(x0 + self.tile_size, self.image.shape[2] - border)
        y1 = min(y0 + self.tile_size, self.image.shape[1] - border)
        best_pos, best_value = _tile_peak(
            y0, x0, y1, x1, self.image, self.clean_parameters.mode,
            self.image.dtype.type(0))
        self._tile_max[y, x] = best_value
        self._tile_pos[y, x] = best_pos

    def _subtract_psf(self, y, x, psf_patch):
        psf_size_x = self.psf.shape[2]
        psf_size_y = self.psf.shape[1]
        psf_patch_x = psf_patch[2]
        psf_patch_y = psf_patch[1]
        image_size_x = self.image.shape[2]
        image_size_y = self.image.shape[1]
        # Centre point to subtract at (x, y) in image
        psf_x = psf_size_x // 2
        psf_y = psf_size_y // 2
        x0 = x - psf_patch_x // 2
        x1 = x0 + psf_patch_x
        y0 = y - psf_patch_y // 2
        y1 = y0 + psf_patch_y
        psf_x0 = psf_x - psf_patch_x // 2
        psf_y0 = psf_y - psf_patch_y // 2
        psf_x1 = psf_x0 + psf_patch_x
        psf_y1 = psf_y0 + psf_patch_y
        if x0 < 0:
            psf_x0 -= x0
            x0 = 0
        if y0 < 0:
            psf_y0 -= y0
            y0 = 0
        if x1 > image_size_x:
            psf_x1 -= (x1 - image_size_x)
            x1 = image_size_x
        if y1 > image_size_y:
            psf_y1 -= (y1 - image_size_y)
            y1 = image_size_y
        scale = self.clean_parameters.loop_gain * self.image[:, y, x]
        self.image[..., y0:y1, x0:x1] -= (
            scale[:, np.newaxis, np.newaxis] * self.psf[..., psf_y0:psf_y1, psf_x0:psf_x1])
        self.model[..., y, x] += scale
        return (y0, x0, y1, x1)

    def reset(self):
        """Call when the dirty image has changed (including before the first
        use) to recompute the internal caches. Note that this does *not* clear
        the model image: the caller is responsible for setting the initial
        values.
        """
        for y in range(self._tile_max.shape[0]):
            for x in range(self._tile_max.shape[1]):
                self._update_tile(y, x)

    def __call__(self, psf_patch, threshold=0.0):
        """Execute a single CLEAN minor cycle."""
        peak_tile = np.unravel_index(np.argmax(self._tile_max), self._tile_max.shape)
        peak_pos = self._tile_pos[peak_tile]
        peak_value = self._tile_max[peak_tile]
        if peak_value < threshold:
            return None
        (y0, x0, y1, x1) = self._subtract_psf(peak_pos[0], peak_pos[1], psf_patch)
        border = self.clean_parameters.border
        tile_y0 = max((y0 - border) // self.tile_size, 0)
        tile_x0 = max((x0 - border) // self.tile_size, 0)
        tile_y1 = min(accel.divup(y1 - border, self.tile_size), self._tile_max.shape[0])
        tile_x1 = min(accel.divup(x1 - border, self.tile_size), self._tile_max.shape[1])
        for y in range(tile_y0, tile_y1):
            for x in range(tile_x0, tile_x1):
                self._update_tile(y, x)
        return peak_value



r"""Fitting of the synthesised beam, and convolution of this beam with the
model.

Convolution implementation details
----------------------------------
This is done by FFT-based convolution, which causes the restoring beam to
wrap at the edges of the image. Since the edges are usually suspect anyway,
this isn't considered an issue.

The Fourier transform of the restoring beam is computed analytically,
rather than applying an FFT to a sampled beam. Once GPU-acceleration is
being done, this will save memory since the transform can be multiplied
in-place as it is computed.

Consider a 1D Gaussian of the form :math:`e^{-\frac12 x^2}`: it has a
Fourier Transform of :math:`\sqrt{2\pi}e^{-2\pi^2 k^2}` (see Mathworld_).
Since a Gaussian is separable, the transform of the 2D Gaussian
:math:`Ae^{-\frac12 \lVert x\rVert^2}` is
:math:`2\pi A e^{-2\pi^2\lVert k\rVert^2}`. Next, using variable
substitution, the transform of
:math:`Ae^{-\frac12 \lVert M^{-1}x\rVert^2}` is
:math:`2\pi \lvert M\rvert A e^{-2\pi^2\lVert Mk\rVert^2}` for a matrix
:math:`M`.

Astropy represents a 2D Gaussian by the standard deviation along the major
and minor axes and the angle of the axes, so we need to reconstruct M,
which is the square root of the covariance matrix. For standard deviations
:math:`\sigma_1` and :math:`\sigma_2` and :math:`R` being the rotation
matrix for the stored angle, the covariance matrix is
:math:`R\left(\begin{smallmatrix}\sigma_1^2 & 0\\0 & \sigma_2^2\end{smallmatrix}\right)R^T`
and its square root is
:math:`M = R\left(\begin{smallmatrix}\sigma_1 & 0\\0 & \sigma_2\end{smallmatrix}\right)R^T`.

.. _Mathworld: http://mathworld.wolfram.com/FourierTransformGaussian.html

.. include:: macros.rst
"""

import math

import numpy as np
import pkg_resources
from astropy.modeling import models, fitting
from astropy import units
from katsdpsigproc import accel
import katsdpimager.types

from katsdpimager import fft


class Beam:
    """Gaussian synthesised beam model.

    Parameters
    ----------
    model : :class:`astropy.modeling.models.Gaussian2D`
        Fitted 2D Gaussian model

    Attributes
    ----------
    model : :class:`astropy.modeling.models.Gaussian2D`
        Model passed to the constructor
    major : float
        Full-width half-maximum for the major axis
    minor : float
        Full-width half-maximum for the minor axis
    theta : angle Quantity
        Angle of the ellipse, measured from positive axis 0 of the PSF to the
        major axis of the ellipse, in the direction of positive axis 1 of the
        PSF.
    """
    def __init__(self, model):
        self.model = model
        # Scale factor between stddev and FWHM
        scale = math.sqrt(8 * math.log(2))
        self.major = model.x_stddev.value * scale
        self.minor = model.y_stddev.value * scale
        theta = model.theta.value
        if self.major < self.minor:
            self.minor, self.major = self.major, self.minor
            theta += math.pi / 2
        theta %= math.pi
        self.theta = theta * units.rad

    def __str__(self):
        return "Beam({0.major}, {0.minor}, {0.theta})".format(self)

    def __repr__(self):
        return "Beam({0.major!r}, {0.minor!r}, {0.theta!r})".format(self)


def fit_beam(psf, step=1.0, threshold=0.01, init_threshold=0.5):
    """Fit a 2D Gaussian to the point spread function. The function is not
    truncated to the central region: the caller should do this.

    Values far from the origin are troublesome because they are not even
    vaguely Gaussian. We restrict the match to values that exceed
    `threshold`.

    Parameters
    ----------
    psf : array-like, 2D and real
        Point spead function, with the origin at the central pixel (rounded up)
    step : float
        Spacing between sample points. The units are arbitrary, and the return
        values will be in the same units.
    threshold : float
        Minimum PSF value to be used in the fit.
    init_threshold : float
        Minimum PSF value used to compute the initial estimate.

    Returns
    -------
    major : float
        Full-width half-maximum for the major axis
    minor : float
        Full-width half-maximum for the minor axis
    theta : float
        Angle of the ellipse, measured from positive axis 0 of the PSF to the
        major axis of the ellipse, in the direction of positive axis 1 of the
        PSF.
    """
    def extract(psf, threshold):
        """Get coordinates and values that are > threshold"""
        mask = psf > threshold
        indices = list(np.nonzero(mask))
        indices[0] = (indices[0] - psf.shape[0] // 2) * step
        indices[1] = (indices[1] - psf.shape[1] // 2) * step
        picked = psf[mask]
        return picked, indices

    # Initial guess: compute the second moment about the origin, but keeping
    # only the values above a (fairly high) threshold.
    picked, indices = extract(psf, init_threshold)
    cov = np.empty((2, 2))
    total = np.sum(picked)
    cov[0, 0] = np.sum(picked * indices[0]**2) / total
    cov[0, 1] = np.sum(picked * indices[0] * indices[1]) / total
    cov[1, 0] = cov[0, 1]
    cov[1, 1] = np.sum(picked * indices[1]**2) / total
    # Because we truncated at init_threshold, we will have an underestimate
    # of the radius. We can correct for this with Mathematics (tm). The
    # variance of a standard 2D Gaussian truncated at radius R is
    # 1 - (1 + R^2/2) * exp(-R*2 / 2).
    R2 = -2 * np.log(init_threshold)
    scale = 1 - (1 + 0.5 * R2) * np.exp(-0.5 * R2)
    cov /= scale
    init = models.Gaussian2D(
        amplitude=1.0,
        cov_matrix=cov,
        fixed={'x_mean': True, 'y_mean': True, 'amplitude': True})

    # Compute a more accurate model by fitting to more of the data
    picked, indices = extract(psf, threshold)
    fit = fitting.LevMarLSQFitter()
    model = fit(init, indices[0], indices[1], picked)
    return Beam(model)


def beam_covariance_sqrt(beam):
    model = beam.model
    # Rotation matrix for theta
    c = np.cos(model.theta)
    s = np.sin(model.theta)
    Q = np.matrix([[c, -s], [s, c]])
    # Diagonalisation of square root of covariance matrix
    D = np.asmatrix(np.diag([model.x_stddev.value, model.y_stddev.value]))
    return Q * D * Q.T


def convolve_beam(model, beam, out=None):
    """Convolve a model image with a restoring beam.

    Parameters
    ----------
    model : array-like, float
        Model image as a 3D array indexed by polarization, m, l
    beam : :class:`Beam`
        Restoring beam model, created with step size equal to the pixel size of
        `model`.
    out : array-like, optional
        If specified, it will contain the output. It is safe to pass `model`.
    """
    if out is None:
        out = np.empty_like(model)
    model_ft = np.fft.fftn(model, axes=[1, 2])
    M = beam_covariance_sqrt(beam)
    # Due to https://github.com/astropy/astropy/issues/1105, the determinant
    # can be negative; hence, the np.abs is necessary.
    amplitude = 2 * np.pi * beam.model.amplitude * np.abs(np.linalg.det(M))
    u = np.fft.fftfreq(model.shape[1])
    v = np.fft.fftfreq(model.shape[2])
    # Coords is shape (model.shape[1], model.shape[2], 2) - pairs of coordinates
    coords = np.stack(np.meshgrid(u, v, indexing='ij'), axis=-1)
    # Compute matrix-vector product M * uv, for every uv coordinate
    rotated_coords = np.inner(coords, M.A)
    rotated_coords_square = np.sum(rotated_coords**2, axis=-1)
    beam_ft = amplitude * np.exp(-2.0 * np.pi**2 * rotated_coords_square)
    out[:] = np.fft.ifftn(model_ft * beam_ft[np.newaxis, ...], axes=[1, 2]).real
    return out


class FourierBeamTemplate:
    """Multiply the Fourier transform of an image by the Fourier transform of a
    Gaussian beam model. The beam parameters must be in units of image pixels.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : {`numpy.float32`, `numpy.float64`}
        Real data type
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, context, dtype, tuning=None):
        # TODO: autotune
        self.wgs_x = 16
        self.wgs_y = 16
        self.dtype = np.dtype(dtype)
        parameters = {
            'wgs_x': self.wgs_x,
            'wgs_y': self.wgs_y,
            'real_type': katsdpimager.types.dtype_to_ctype(self.dtype)
        }
        self.program = accel.build(
            context, "imager_kernels/fourier_beam.mako", parameters,
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return FourierBeam(self, *args, **kwargs)


class FourierBeam(accel.Operation):
    """Instantiation of :class:`FourierBeamTemplate`.

    .. rubric:: Slots

    **data** : array of shape (height, width // 2 + 1), complex
        Fourier transform (real-to-complex) of the model image. On output, the
        Fourier transform of the restored image.

    Attributes
    ----------
    beam : :class:`Beam`
        Restoring beam. It must be set before invoking the operation.

    Parameters
    ----------
    template : :class:`FourierBeamTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    image_shape : tuple
        (height, width) of the image. Note that the buffer on which this operation works
        will have shape (height, width // 2 + 1).
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, image_shape, allocator=None):
        if len(image_shape) != 2:
            raise ValueError('image_shape must be 2D')
        super().__init__(command_queue, allocator=allocator)
        self.template = template
        self.image_shape = image_shape
        output_shape = (image_shape[0], image_shape[1] // 2 + 1)
        complex_type = katsdpimager.types.real_to_complex(template.dtype)
        self.slots['data'] = accel.IOSlot(output_shape, complex_type)
        self.beam = None
        self._kernel = template.program.get_kernel('fourier_beam')

    def _run(self):
        if self.beam is None:
            raise ValueError('Must set beam')
        M = beam_covariance_sqrt(self.beam)
        # Due to https://github.com/astropy/astropy/issues/1105, the determinant
        # can be negative; hence, the np.abs is necessary.
        amplitude = 2 * np.pi * self.beam.model.amplitude * np.abs(np.linalg.det(M))
        # CUFFT, unlikely numpy.fft, does not normalize the inverse transform.
        # We fold the normalization into the amplitude
        amplitude /= self.image_shape[0] * self.image_shape[1]
        # The kernel has integral coordinates, which we need to convert to
        # the range [-1, 1). We fold this into the matrix.
        M *= np.asmatrix(np.diag([1.0 / self.image_shape[0],
                                  1.0 / self.image_shape[1]]))
        # Compute overall matrix for the exponent
        C = -2 * np.pi**2 * M.T * M
        real = self.template.dtype.type
        data = self.buffer('data')
        self.command_queue.enqueue_kernel(
            self._kernel,
            [
                data.buffer,
                np.int32(data.padded_shape[1]),
                real(amplitude),
                real(C[0, 0]),
                real(2 * C[0, 1]),
                real(C[1, 1]),
                np.int32(data.shape[1]),
                np.int32(data.shape[0])
            ],
            global_size=(accel.roundup(data.shape[1], self.template.wgs_x),
                         accel.roundup(data.shape[0], self.template.wgs_y)),
            local_size=(self.template.wgs_x, self.template.wgs_y)
        )


class ConvolveBeamTemplate:
    """Template for device convolution of a model image with a Gaussian
    restoring beam, for a single polarization. Due to limitations in CUFFT,
    many parameters are folded into the template.

    Due to limitations in katsdpsigproc, convolving a multi-polarization image
    currently requires each polarization to be copied into a
    single-polarization image first.

    Parameters
    ----------
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple
        (height, width) of the image
    dtype : {`numpy.float32`, `numpy.float64`}
        Type of the image data
    padded_shape_image : tuple, optional
        Padded shape for the image. Defaults to no padding.
    padded_shape_fourier : tuple, optional
        Padded shape for the Fourier transform of the image. Defaults to no
        padding. Note that because the image is real, the Fourier transform has
        shape (height, width // 2 + 1).
    tuning : dict, optional
        Tuning parameters (currently unused)
    """
    def __init__(self, command_queue, shape, dtype,
                 padded_shape_image=None, padded_shape_fourier=None, tuning=None):
        if padded_shape_image is None:
            padded_shape_image = tuple(shape)
        if padded_shape_fourier is None:
            padded_shape_fourier = tuple(shape[:-1]) + (shape[-1] // 2 + 1,)
        if len(shape) != 2 or len(padded_shape_image) != 2 or len(padded_shape_fourier) != 2:
            raise ValueError('wrong number of dimensions')
        self.dtype = np.dtype(dtype)
        self.shape = shape
        complex_dtype = katsdpimager.types.real_to_complex(self.dtype)
        self.fft = fft.FftTemplate(command_queue, 2, shape, dtype, complex_dtype,
                                   padded_shape_image, padded_shape_fourier)
        self.ifft = fft.FftTemplate(command_queue, 2, shape, complex_dtype, dtype,
                                    padded_shape_fourier, padded_shape_image)
        self.fourier_beam = FourierBeamTemplate(command_queue.context, dtype, tuning)

    def instantiate(self, *args, **kwargs):
        return ConvolveBeam(self, *args, **kwargs)


class ConvolveBeam(accel.OperationSequence):
    """Instantiation of :class:`ConvolveBeamTemplate`.

    .. rubric:: Slots

    **image** : array of shape (height, width), real
        Image to be convolved, as well as the output.
    **fourier** : array of shape (height, width // 2 + 1), real
        Fourier transform of the image. It should be treated as an opaque
        scratch buffer.

    Attributes
    ----------
    beam : :class:`Beam`
        Restoring beam. It must be set before invoking the operation.
    """
    def __init__(self, template, allocator=None):
        self._fft = template.fft.instantiate(fft.FFT_FORWARD, allocator=allocator)
        self._ifft = template.ifft.instantiate(fft.FFT_INVERSE, allocator=allocator)
        self._fourier_beam = template.fourier_beam.instantiate(
            template.fft.command_queue, template.shape, allocator=allocator)
        operations = [
            ('fft', self._fft),
            ('fourier_beam', self._fourier_beam),
            ('ifft', self._ifft)
        ]
        compounds = {
            'image': ['fft:src', 'ifft:dest'],
            'fourier': ['fft:dest', 'ifft:src', 'fourier_beam:data']
        }
        super().__init__(
            template.fft.command_queue, operations, compounds, allocator=allocator)

    @property
    def beam(self):
        return self._fourier_beam.beam

    @beam.setter
    def beam(self, value):
        self._fourier_beam.beam = value



#!/usr/bin/env python

"""Compare two FITS images, applying WCS transformations so that comparisons
are at points with the same world coordinates rather than blindly matching
pixels.

At present, only single-channel images and only the primary HDU are supported.

The operation is asymmetric: for each pixel in the first file, we look up the
corresponding pixel in the second file, interpolating as necessary, or
reporting NaN if it falls outside the footprint.
"""

import argparse

import numpy as np
import scipy.interpolate
import astropy.io.fits as fits
import astropy.wcs as wcs


class FitsFile:
    def __init__(self, filename):
        self.hdulist = fits.open(filename)
        self.xform = wcs.WCS(self.hdulist[0].header, self.hdulist)
        self.data = self.hdulist[0].data
        self.axismap = [None, None, None]   # RA, DEC, Stokes
        for i, t in enumerate(self.xform.get_axis_types()):
            if t['coordinate_type'] == 'celestial' and t['number'] in [0, 1]:
                self.axismap[t['number']] = i
            elif t['coordinate_type'] == 'stokes':
                self.axismap[2] = i
        if None in self.axismap:
            raise ValueError('Not all required axes found')

    def to_world(self, coords):
        # WCS takes indices in Fortran order
        world = self.xform.all_pix2world(coords[..., ::-1], 0)
        return world[:, self.axismap]

    def from_world(self, world):
        mid = np.zeros((world.shape[0], len(self.xform.axis_type_names)), np.float_)
        mid[:, self.axismap] = world
        # WCS returns indices in Fortran order
        return self.xform.all_world2pix(mid, 0)[..., ::-1]


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('file', nargs=3)
    args = parser.parse_args()

    files = [FitsFile(filename) for filename in args.file[:2]]
    coords0 = np.indices(files[0].data.shape)
    # Rearrange structure of arrays of indices to array of multi-indices
    coords0 = np.rollaxis(coords0, 0, len(coords0.shape))
    # Flatten the multi-dimensional array created by np.indices
    coords0 = coords0.reshape((-1, coords0.shape[-1]))
    world = files[0].to_world(coords0)
    coords1 = files[1].from_world(world)
    # Strip out axes with size 1, since they cause interpn issues
    keep_axes = [x[0] for x in enumerate(files[1].data.shape) if x[1] > 1]
    squeezed = np.squeeze(files[1].data)
    values1 = scipy.interpolate.interpn(
        tuple(np.arange(x) for x in squeezed.shape),
        squeezed,
        coords1[..., keep_axes],
        bounds_error=False)
    # Unflatten the data again
    values1 = values1.reshape(files[0].data.shape)
    delta = values1 - files[0].data

    header = files[0].hdulist[0].header.copy()
    header['ORIGIN'] = 'fitsdiffwcs'
    hdu = fits.PrimaryHDU(delta, header)
    hdu.writeto(args.file[2], overwrite=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import tempfile
from unittest import mock

from nose.tools import assert_equal, assert_raises
import fakeredis
from astropy import units
import numpy as np
import katpoint
import katsdptelstate.redis

from katsdpimager.sky_model import (KatpointSkyModel, catalogue_from_telstate, open_sky_model,
                                    NoSkyModelError)


_TRG_A = 'A, radec, 20:00:00.00, -60:00:00.0, (200.0 12000.0 1.0 0.5)'
_TRG_B = 'B, radec, 8:00:00.00, 60:00:00.0, (200.0 12000.0 2.0)'
_TRG_C = 'C, radec, 21:00:00.00, -60:00:00.0, (800.0 43200.0 1.0 0.0 0.0 0.0 0.0 0.0 1.0 0.8 -0.7 0.6)'    # noqa: E501


class TestKatpointSkyModel:
    def setUp(self):
        self.catalogue = katpoint.Catalogue([_TRG_A, _TRG_B, _TRG_C])
        self.sky_model = KatpointSkyModel(self.catalogue)

    def test_len(self):
        assert_equal(len(self.sky_model), 3)

    def test_lmn(self):
        phase_centre = np.array([300, -60]) * units.deg   # RA 20.0
        lmn = self.sky_model.lmn(phase_centre)
        # The LMN coordinates for C were determined experimentally
        np.testing.assert_allclose(lmn, [
            [0, 0, 1],
            [0, 0, -1],
            [1.294095e-01, -1.475455e-02, 9.914815e-01]], atol=1e-5)

    def test_flux_density(self):
        flux = self.sky_model.flux_density(1e10 * units.Hz)
        np.testing.assert_allclose(flux, [
            [1000, 0, 0, 0],
            [100, 0, 0, 0],
            [10, 8, -7, 6]])

        # Check that frequency outside the range of the flux density model returns zeros
        flux = self.sky_model.flux_density(500e6 * units.Hz)
        np.testing.assert_allclose(flux, [
            [223.606798, 0, 0, 0],
            [100, 0, 0, 0],
            [0, 0, 0, 0]])


def _put_models(telstate, cbid, continuum, models):
    """Populate telstate with continuum models.

    Each element of `models` is a tuple containing a phase centre and a list of
    components. The component list may also be ``None`` to omit that part of
    the value (to test failure handling).
    """
    # Uses bytes strings to emulate katsdpcontim (which uses Python 2 native strings)
    targets = {}
    for i, (phase_centre, components) in enumerate(models):
        normalised_name = 'test{}'.format(i)
        d = {b'description': phase_centre.encode('utf-8')}
        if components is not None:
            d[b'components'] = [component.encode('utf-8') for component in components]
        namespace = telstate.join(cbid, continuum, normalised_name, 'target0')
        telstate.view(namespace)['clean_components'] = d
        targets[phase_centre] = normalised_name
    telstate.view(telstate.join(cbid, continuum))['targets'] = targets


class TestCatalogueFromTelstate:
    def setup(self):
        self.cbid = '1234567890'
        self.continuum = 'continuum'
        self.telstate = katsdptelstate.TelescopeState()
        self.telstate['sdp_archived_streams'] = [self.continuum]
        self.telstate[self.telstate.join(self.continuum, 'stream_type')] = 'sdp.continuum_image'
        _put_models(self.telstate, self.cbid, self.continuum, [
            (_TRG_A, [_TRG_A, _TRG_C]),
            (_TRG_B, [_TRG_B]),
            ('Nothing, special', None)
        ])

    def test_present(self):
        catalogue = catalogue_from_telstate(self.telstate, self.cbid, self.continuum,
                                            katpoint.Target(_TRG_A))
        assert_equal(len(catalogue), 2)
        assert_equal(catalogue.targets[0], katpoint.Target(_TRG_A))
        assert_equal(catalogue.targets[1], katpoint.Target(_TRG_C))

        catalogue = catalogue_from_telstate(self.telstate, self.cbid, self.continuum,
                                            katpoint.Target(_TRG_B))
        assert_equal(len(catalogue), 1)
        assert_equal(catalogue.targets[0], katpoint.Target(_TRG_B))

    def test_auto_continuum(self):
        self.continuum = None
        self.test_present()

    def test_absent(self):
        with assert_raises(NoSkyModelError):
            catalogue_from_telstate(self.telstate, self.cbid, self.continuum,
                                    katpoint.Target(_TRG_C))

    def test_missing_key(self):
        with assert_raises(NoSkyModelError):
            catalogue_from_telstate(self.telstate, self.cbid, self.continuum,
                                    katpoint.Target('Nothing, special'))


class TestOpenSkyModel:
    def test_bad_format(self):
        with assert_raises(ValueError):
            open_sky_model('file:///does_not_exist?format=sir_not_appearing_in_this_codebase')

    def test_bad_scheme(self):
        with assert_raises(ValueError):
            open_sky_model('ftp://invalid/')

    def test_missing_params(self):
        with assert_raises(ValueError):
            open_sky_model('redis://invalid/?capture_block_id=1234567890'
                           '&continuum=continuum&format=katdal')

    def test_file(self):
        orig = katpoint.Catalogue([_TRG_A, _TRG_B, _TRG_C])
        with tempfile.NamedTemporaryFile('w', suffix='.csv') as f:
            orig.save(f.name)
            test1 = open_sky_model(f.name)
            test2 = open_sky_model('file://' + f.name + '?format=katpoint')
        assert_equal(orig, test1._catalogue)
        assert_equal(orig, test2._catalogue)

    def test_telstate(self):
        client = fakeredis.FakeStrictRedis()
        telstate = katsdptelstate.TelescopeState(katsdptelstate.redis.RedisBackend(client))
        # Fake just enough of telstate to keep katdal happy. This isn't all in the right
        # namespace, but that doesn't really matter.
        telstate['stream_name'] = 'sdp_l0'
        telstate_l0 = telstate.view('sdp_l0')
        telstate_l0['stream_type'] = 'sdp.vis'
        telstate_l0['chunk_info'] = {
            'correlator_data': {
                'prefix': '1234567890-sdp-l0',
                'dtype': '<c8',
                'shape': (0, 0, 0)
            }
        }
        telstate_l0['sync_time'] = 1234567890.0
        telstate_l0['first_timestamp'] = 0.0
        telstate_l0['int_time'] = 1.0

        _put_models(telstate, '1234567890', 'continuum', [(_TRG_A, [_TRG_A, _TRG_C])])
        expected = katpoint.Catalogue([_TRG_A, _TRG_C])

        with mock.patch('redis.StrictRedis', return_value=client) as mock_redis:
            test = open_sky_model(
                'redis://invalid:6379/?format=katdal&db=1&capture_block_id=1234567890'
                '&continuum=continuum'
                '&target=A,+radec,+20:00:00.00,+-60:00:00.0,+(200.0+12000.0+1.0+0.5+0.0)')
            mock_redis.assert_called_with(host='invalid', port=6379, db=1, socket_timeout=mock.ANY)
            assert_equal(expected, test._catalogue)



# -*- coding: utf-8 -*-

r"""
Computation of density weights, for uniform and robust weighting. Refer to
[Bri95]_ for definitions. Computation of weights is done in up to four
stages:

1. The statistical weights are gridded, *without* any convolution. The subgrid
   coordinates are still passed so that the same data layouts can be shared with
   convolutional gridding, but they are be ignored.

2. For robust weighting, an "average" statistical weight is computed as

   .. math:: \overline{W} = \frac{\sum W_i^2}{\sum W_i},

   where the sums are over grid cells.

3. The total statistical weight per cell is converted into a density weight.
   If :math:`W_i` is the total statistical weight for a cell, then a uniform
   weight is computed as :math:`1 / W_i` and a robust weight as
   :math:`1 / (W_i S^2 + 1)`, where

   .. math:: S^2 = \overline{W}(5\cdot 10^{-R})^2

   and :math:`R` is the robustness parameter. The conversion formula for robust
   weighting is taken from wsclean.

4. During gridding, as visibilities are loaded the density weights are looked
   up and multiplied in. The visibilities are already pre-weighted by the
   statistical weights.

Note that all these steps use compressed visibilities. This works because the
density weights are constant for a grid cell, and thus constant across the
original visibilities that contribute to a compressed visibility.

Weights are processed separately per polarization. However, the
robustness parameter :math:`S` is computed for the first polarization
(generally Stokes I) and used for all polarizations, to avoid qualitatively
different beam shapes for the different polarizations.

.. [Bri95] Briggs, D. S. 1995. High fidelity deconvolution of moderately
   resolved sources. PhD Thesis, The New Mexico Institute of Mining and Technology.
   http://www.aoc.nrao.edu/dissertations/dbriggs/

.. include:: macros.rst
"""

import enum

import pkg_resources
import numpy as np
from katsdpsigproc import accel, fill


class WeightType(enum.Enum):
    NATURAL = 0
    UNIFORM = 1
    ROBUST = 2


class GridWeightsTemplate:
    """Template for accumulating weights onto a grid.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    num_polarizations : int
        Number of polarizations to grid
    tuning : dict, optional
        Tuning parameters (unused)
    """
    def __init__(self, context, num_polarizations, tuning=None):
        self.context = context
        self.num_polarizations = num_polarizations
        # TODO: autotuning
        self.wgs = 256
        parameters = {
            'num_polarizations': num_polarizations,
            'wgs': self.wgs
        }
        self.program = accel.build(
            context, "imager_kernels/grid_weights.mako", parameters,
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return GridWeights(self, *args, **kwargs)


class GridWeights(accel.Operation):
    """Instantiation of :class:`GridWeightsTemplate`.

    The user should set some prefix of the `uv` and `weights` slots, set
    :attr:`num_vis` to indicate how many weights are provided, and then call
    the operation to grid them.

    .. rubric:: Slots

    **uv** : array of shape `max_vis` × 4, int16
        UV coordinates of the grid points. The coordinates are biased by half
        the grid size, so that (0, 0) refers to the centre of the grid.
    **weights** : array of shape `max_vis` × `num_polarizations`, float32
        Weights to accumulate
    **grid** : array of shape `num_polarizations` × height × width, float32
        Accumulated weights

    Parameters
    ----------
    template : :class:`GridWeightsTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    grid_shape : tuple of ints
        Shape for the grid, (polarizations, height, width)
    max_vis : int
        Maximum number of weights that can be gridded in one pass
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Raises
    ------
    ValueError
        if the grid width or height is odd, or if the number of polarizations
        does not match `template`.
    """
    def __init__(self, template, command_queue, grid_shape, max_vis, allocator=None):
        super().__init__(command_queue, allocator)
        self.template = template
        if grid_shape[0] != self.template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        if grid_shape[1] % 2 or grid_shape[2] % 2:
            raise ValueError('Odd-sized grid not currently supported')
        self.max_vis = max_vis
        self.slots['grid'] = accel.IOSlot(grid_shape, np.float32)
        self.slots['uv'] = accel.IOSlot(
            (max_vis, accel.Dimension(4, exact=True)), np.int16)
        self.slots['weights'] = accel.IOSlot(
            (max_vis, accel.Dimension(self.template.num_polarizations, exact=True)), np.float32)
        self._num_vis = 0
        self._kernel = self.template.program.get_kernel('grid_weights')

    @property
    def num_vis(self):
        """The actual number of weights to grid"""
        return self._num_vis

    @num_vis.setter
    def num_vis(self, n):
        """Change the number of actual visibilities stored in the buffers."""
        if n < 0 or n > self.max_vis:
            raise ValueError('Number of visibilities {} is out of range 0..{}'.format(
                n, self.max_vis))
        self._num_vis = n

    def _run(self):
        grid = self.buffer('grid')
        uv = self.buffer('uv')
        weights = self.buffer('weights')
        half_v = grid.shape[1] // 2
        half_u = grid.shape[2] // 2
        address_bias = half_v * grid.padded_shape[2] + half_u
        self.command_queue.enqueue_kernel(
            self._kernel,
            [
                grid.buffer,
                np.int32(grid.padded_shape[2]),
                np.int32(grid.padded_shape[1] * grid.padded_shape[2]),
                uv.buffer,
                weights.buffer,
                np.int32(address_bias),
                np.int32(self._num_vis)
            ],
            global_size=(accel.roundup(self._num_vis, self.template.wgs),),
            local_size=(self.template.wgs,)
        )

    def parameters(self):
        return {
            'wgs': self.template.wgs,
            'num_polarizations': self.template.num_polarizations,
            'max_vis': self.max_vis
        }


class DensityWeightsTemplate:
    """Template for converting cell sum of statistical weights to density weights.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    num_polarizations : int
        Number of polarizations to grid
    tuning : dict, optional
        Tuning parameters (unused)
    """
    def __init__(self, context, num_polarizations, tuning=None):
        self.context = context
        self.num_polarizations = num_polarizations
        # TODO: autotuning
        self.wgs_x = 16
        self.wgs_y = 16
        parameters = {
            'num_polarizations': num_polarizations,
            'wgs_x': self.wgs_x,
            'wgs_y': self.wgs_y
        }
        self.program = accel.build(
            context, "imager_kernels/density_weights.mako", parameters,
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return DensityWeights(self, *args, **kwargs)


class DensityWeights(accel.Operation):
    """Instantiation of :class:`DensityWeightsTemplate`. This operation
    modifies the grid in-place. The type of weights is controlled by two
    attributes, :attr:`a` and :attr:`b`. For an input :math:`W`, the result
    is :math:`1 / (aW + b)`. This allows for natural, uniform or robust
    weights. If not altered, the default gives uniform weights.

    It also computes and returns the normalized thermal RMS, as described in
    equation 3.5 of [Bri95]_.

    .. rubric:: Slots

    **grid** : array of shape `num_polarizations` × height × width, float32
        On input, the sum of statistical weights per cell; on output, the density
        weight for each cell.

    Parameters
    ----------
    template : :class:`DensityWeightsTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    grid_shape : tuple of ints
        Shape for the grid, (polarizations, height, width)
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, grid_shape, allocator=None):
        super().__init__(command_queue, allocator)
        self.template = template
        if grid_shape[0] != self.template.num_polarizations:
            raise ValueError('Mismatch in number of polarizations')
        self.a = 1.0
        self.b = 0.0
        self.slots['grid'] = accel.IOSlot(
            (grid_shape[0],
             accel.Dimension(grid_shape[1], self.template.wgs_y),
             accel.Dimension(grid_shape[2], self.template.wgs_x)),
            np.float32)
        self.slots['sums'] = accel.IOSlot((3,), np.float32)
        self._kernel = self.template.program.get_kernel('density_weights')
        self._sums_host = accel.HostArray((3,), np.float32, context=self.template.context)

    def _run(self):
        grid = self.buffer('grid')
        sums = self.buffer('sums')
        sums.zero(self.command_queue)
        self.command_queue.enqueue_kernel(
            self._kernel,
            [
                sums.buffer,
                grid.buffer,
                np.int32(grid.padded_shape[2]),
                np.int32(grid.padded_shape[1] * grid.padded_shape[2]),
                np.int32(grid.shape[2]),
                np.int32(grid.shape[1]),
                np.float32(self.a),
                np.float32(self.b)
            ],
            global_size=(accel.roundup(grid.shape[2], self.template.wgs_x),
                         accel.roundup(grid.shape[1], self.template.wgs_y)),
            local_size=(self.template.wgs_x, self.template.wgs_y)
        )
        if isinstance(sums, accel.SVMArray):
            self.command_queue.finish()
        sums.get(self.command_queue, self._sums_host)
        return np.sqrt(self._sums_host[2] * self._sums_host[0]) / self._sums_host[1]

    def parameters(self):
        return {
            'a': self.a,
            'b': self.b,
            'wgs_x': self.template.wgs_x,
            'wgs_y': self.template.wgs_y,
            'num_polarizations': self.template.num_polarizations,
        }


class MeanWeightTemplate:
    """Template for computing the "mean weight", as defined by equation 3.17 in
    [Bri95]_. The input is the gridded statistical weights. The kernel outputs the
    sum of squared cell weights and the sum of weights, and the Python code
    computes the mean weight from these. Only the first polarization is used.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    tuning : dict, optional
        Tuning parameters (unused)
    """
    def __init__(self, context, tuning=None):
        self.context = context
        # TODO: autotuning
        self.wgs_x = 16
        self.wgs_y = 16
        parameters = {
            'wgs_x': self.wgs_x,
            'wgs_y': self.wgs_y
        }
        self.program = accel.build(
            context, "imager_kernels/mean_weight.mako", parameters,
            extra_dirs=[pkg_resources.resource_filename(__name__, '')])

    def instantiate(self, *args, **kwargs):
        return MeanWeight(self, *args, **kwargs)


class MeanWeight(accel.Operation):
    """Instantiation of :class:`MeanWeightTemplate`.

    .. rubric:: Slots

    **grid** : array of shape `num_polarizations` × height × width, float32
        The sum of statistical weights per cell

    Parameters
    ----------
    template : :class:`MeanWeightTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    grid_shape : tuple of ints
        Shape for the grid, (polarizations, height, width)
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, grid_shape, allocator=None):
        super().__init__(command_queue, allocator)
        self.template = template
        self.slots['grid'] = accel.IOSlot(
            (grid_shape[0],
             accel.Dimension(grid_shape[1], self.template.wgs_y),
             accel.Dimension(grid_shape[2], self.template.wgs_x)),
            np.float32)
        self.slots['sums'] = accel.IOSlot((2,), np.float32)
        self._kernel = template.program.get_kernel('mean_weight')
        self._sums_host = accel.HostArray((2,), np.float32, context=self.template.context)

    def _run(self):
        grid = self.buffer('grid')
        sums = self.buffer('sums')
        self.command_queue.enqueue_zero_buffer(sums.buffer)
        self.command_queue.enqueue_kernel(
            self._kernel,
            [
                sums.buffer,
                grid.buffer,
                np.int32(grid.padded_shape[-1]),
                np.int32(grid.shape[2]),
                np.int32(grid.shape[1])
            ],
            global_size=(accel.roundup(grid.shape[2], self.template.wgs_x),
                         accel.roundup(grid.shape[1], self.template.wgs_y)),
            local_size=(self.template.wgs_x, self.template.wgs_y)
        )
        if isinstance(sums, accel.SVMArray):
            self.command_queue.finish()
        sums.get(self.command_queue, self._sums_host)
        return self._sums_host[1] / self._sums_host[0]


class WeightsTemplate:
    """Compound template for computing imaging weights.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    weight_type : :class:`WeightType`
        Weighting method
    num_polarizations : int
        Number of polarizations
    grid_weights_tuning, mean_weights_tuning, density_weights_tuning : dict, optional
        Tuning parameters passed to :class:`GridWeightsTemplate`,
        :class:`MeanWeightsTemplate` and :class:`DensityWeightsTemplate` respectively
    """
    def __init__(self, context, weight_type, num_polarizations,
                 grid_weights_tuning=None, mean_weight_tuning=None,
                 density_weights_tuning=None):
        self.context = context
        self.weight_type = weight_type
        if weight_type == WeightType.NATURAL:
            self.grid_weights = None
            self.mean_weight = None
            self.density_weights = None
            self.fill = fill.FillTemplate(context, np.float32, 'float')
        else:
            self.grid_weights = GridWeightsTemplate(context, num_polarizations,
                                                    tuning=grid_weights_tuning)
            if weight_type == WeightType.ROBUST:
                self.mean_weight = MeanWeightTemplate(context, tuning=mean_weight_tuning)
            else:
                self.mean_weight = None
            self.density_weights = DensityWeightsTemplate(context, num_polarizations,
                                                          tuning=density_weights_tuning)
            self.fill = None

    def instantiate(self, *args, **kwargs):
        return Weights(self, *args, **kwargs)


class Weights(accel.OperationSequence):
    """Instantiation of :class:`WeightsTemplate`. The steps to use it are:

     1. If using :const:`WeightType.ROBUST` weighting, set :attr:`robustness`.
     2. Call :meth:`clear`.
     3. For each batch of weights, call :meth:`grid`.
     4. Call :meth:`finalize`, which returns the normalized thermal RMS.

    .. rubric:: Slots

    **uv** : array of shape `max_vis` × 4, int16
        UV coordinates of the grid points. The coordinates are biased by half
        the grid size, so that (0, 0) refers to the centre of the grid.
    **weights** : array of shape `max_vis` × `num_polarizations`, float32
        Weights to accumulate
    **grid** : array of shape `num_polarizations` × height × width, float32
        Accumulated weights

    .. note::

        The **uv** and **weights** slots will be absent if the template was
        created with :const:`WeightType.NATURAL` weighting.

    Parameters
    ----------
    command_queue : |CommandQueue|
        Command queue for the operation
    grid_shape : tuple of ints
        Shape for the grid, (polarizations, height, width)
    max_vis : int
        Maximum number of weights that can be gridded in one pass
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Attributes
    ----------
    robustness : float
        Parameter for robust weighting
    """
    def __init__(self, template, command_queue, grid_shape, max_vis, allocator=None):
        self.template = template
        operations = []
        compounds = {'grid': []}

        if template.grid_weights is not None:
            self._grid_weights = template.grid_weights.instantiate(
                command_queue, grid_shape, max_vis, allocator)
            operations.append(('grid_weights', self._grid_weights))
            compounds['grid'].append('grid_weights:grid')
            compounds['uv'] = ['grid_weights:uv']
            compounds['weights'] = ['grid_weights:weights']
        else:
            self._grid_weights = None

        if template.fill is not None:
            self._fill = template.fill.instantiate(
                command_queue, grid_shape, allocator)
            self._fill.set_value(1)
            operations.append(('fill', self._fill))
            compounds['grid'].append('fill:data')
        else:
            self._fill = None

        if template.mean_weight is not None:
            self._mean_weight = template.mean_weight.instantiate(
                command_queue, grid_shape, allocator)
            operations.append(('mean_weight', self._mean_weight))
            compounds['grid'].append(('mean_weight:grid'))
            self.robustness = 0.0
        else:
            self._mean_weight = None
            self.robustness = None

        if template.density_weights is not None:
            self._density_weights = template.density_weights.instantiate(
                command_queue, grid_shape, allocator)
            operations.append(('density_weights', self._density_weights))
            compounds['grid'].append(('density_weights:grid'))
        else:
            self._density_weights = None

        super().__init__(command_queue, operations, compounds, allocator=allocator)

    def _run(self):
        raise NotImplementedError('Weights should not be used as a callable')

    def clear(self):
        self.ensure_all_bound()
        # If self._fill is set, it is not necessary to clear first because
        # finalize will set all relevant values.
        if self._fill is None:
            self.buffer('grid').zero(self.command_queue)

    def grid(self, uv, weights):
        self.ensure_all_bound()
        if self._grid_weights is not None:
            N = len(uv)
            self._grid_weights.num_vis = N
            self.buffer('uv')[:N, 0:2] = uv
            self.buffer('weights')[:N] = weights
            return self._grid_weights()

    def finalize(self):
        self.ensure_all_bound()
        if self._mean_weight is not None:
            # It must be robust weighting
            mean_weight = self._mean_weight()
            S2 = (5 * 10**(-self.robustness))**2 / mean_weight
            self._density_weights.a = S2
            self._density_weights.b = 1.0
        if self._density_weights is not None:
            normalized_rms = self._density_weights()
        else:
            normalized_rms = 1.0
        if self._fill is not None:
            self._fill()
        return normalized_rms


class WeightsHost:
    """Equivalent to :class:`Weights` that runs on the host.

    Parameters
    ----------
    weight_type : :class:`WeightType`
        Weighting method
    weights_grid : array-like, float
        Grid for weights

    Attributes
    ----------
    robustness : float
        Parameter for robust weighting
    """
    def __init__(self, weight_type, weights_grid):
        self.weight_type = weight_type
        self.robustness = 0.0
        self.weights_grid = weights_grid
        assert weights_grid.shape[1] % 2 == 0 and weights_grid.shape[2] % 2 == 0, \
            "Only even-sized grids are currently supported"

    def clear(self):
        if self.weight_type != WeightType.NATURAL:
            self.weights_grid.fill(0)

    def grid(self, uv, weights):
        shape = self.weights_grid.shape
        # Bias uv coordinates to grid center
        uv += np.array([[shape[2] // 2, shape[1] // 2]], np.int16)
        for i in range(len(uv)):
            self.weights_grid[:, uv[i, 1], uv[i, 0]] += weights[i, :]

    def finalize(self):
        if self.weight_type == WeightType.NATURAL:
            self.weights_grid.fill(1)
            normalized_rms = 1.0
        elif self.weight_type == WeightType.UNIFORM:
            sum_w = np.sum(self.weights_grid[0])
            sum_dw = np.count_nonzero(self.weights_grid[0])
            # Force density weights to be zero for cells with no visibilities
            self.weights_grid[self.weights_grid == 0] = np.inf
            np.reciprocal(self.weights_grid, out=self.weights_grid)
            # d^2w == d, because d is 1/w
            sum_d2w = np.sum(self.weights_grid[0])
            normalized_rms = np.sqrt(sum_d2w * sum_w) / sum_dw
        elif self.weight_type == WeightType.ROBUST:
            sum_sq = np.dot(self.weights_grid[0].flat, self.weights_grid[0].flat)
            sum = np.sum(self.weights_grid[0])
            mean_weight = sum_sq / sum
            S2 = (5 * 10**(-self.robustness))**2 / mean_weight
            old_weights0 = self.weights_grid[0].copy()
            # Force density weights to be zero for cells with no visibilities
            self.weights_grid[self.weights_grid == 0] = np.inf
            np.reciprocal(self.weights_grid * S2 + 1, out=self.weights_grid)
            sum_w = np.sum(old_weights0)
            sum_dw = np.sum(self.weights_grid[0] * old_weights0)
            sum_d2w = np.sum(self.weights_grid[0]**2 * old_weights0)
            normalized_rms = np.sqrt(sum_d2w * sum_w) / sum_dw
        else:
            raise ValueError('Unknown weight_type {}'.format(self.weight_type))
        return normalized_rms



"""Routines for dealing with transformations between polarization bases

Enumeration of polarizations uses the CASA enumeration values, which are
different from those used in FITS:

.. py:data:: STOKES_I

.. py:data:: STOKES_Q

.. py:data:: STOKES_U

.. py:data:: STOKES_V

.. py:data:: STOKES_RR

.. py:data:: STOKES_RL

.. py:data:: STOKES_LR

.. py:data:: STOKES_LL

.. py:data:: STOKES_XX

.. py:data:: STOKES_XY

.. py:data:: STOKES_YX

.. py:data:: STOKES_YY
"""

import numpy as np


STOKES_I = 1
STOKES_Q = 2
STOKES_U = 3
STOKES_V = 4
STOKES_RR = 5
STOKES_RL = 6
STOKES_LR = 7
STOKES_LL = 8
STOKES_XX = 9
STOKES_XY = 10
STOKES_YX = 11
STOKES_YY = 12

STOKES_IQUV = [STOKES_I, STOKES_Q, STOKES_U, STOKES_V]

#: Names for polarizations used in display and command line
STOKES_NAMES = [None, 'I', 'Q', 'U', 'V', 'RR', 'RL', 'LR', 'LL', 'XX', 'XY', 'YX', 'YY']

#: Coefficients for each polarization relative to IQUV
STOKES_COEFF = np.array([
    [0, 0, 0, 0],
    [1, 0, 0, 0],
    [0, 1, 0, 0],
    [0, 0, 1, 0],
    [0, 0, 0, 1],
    [1, 0, 0, 1],
    [0, 1, 1j, 0],
    [0, 1, -1j, 0],
    [1, 0, 0, -1],
    [1, 1, 0, 0],
    [0, 0, 1, 1j],
    [0, 0, 1, -1j],
    [1, -1, 0, 0]], np.complex64)


def polarization_matrix(outputs, inputs):
    """Return a Mueller matrix that will map the input polarizations to the outputs.

    The matrix is computed using linear algebra. Let `s` represent the Stokes
    parameters (IQUV). The inputs correspond to :math:`As`, for some matrix
    `A`, and similarly the outputs to :math:`Bs`. We are thus searching for
    the matrix `X` such that :math:`Bs=XAs` or :math:`A^TX^T = B^T`.

    If redundant inputs are given, then the solution is not uniquely
    determined.

    This function is general, and does not require 4 inputs and outputs, or
    even the same number of inputs and outputs.

    Raises
    ------
    ValueError
        if the inputs do not contain the necessary elements to compute the
        outputs.
    """
    A = np.matrix(STOKES_COEFF[inputs, :]).T
    B = np.matrix(STOKES_COEFF[outputs, :]).T
    X, residuals, rank, s = np.linalg.lstsq(A, B, 1e-5)
    # We can't just check for non-zero residuals, because lstsq doesn't
    # return them if A is rank-deficient. Rank-deficiency doesn't
    # necessarily mean there isn't a solution, if B lies in the subspace
    # spanned by A.
    if np.linalg.norm(A * X - B, 'fro') > 1e-5:
        raise ValueError('no solution')
    # In the common cases, the values are all multiples of 0.25, but lstsq has
    # rounding errors. Round off anything that is close enough to a multiple.
    # In particular, tiny values will be flushed to zero, which is important
    # to apply_mueller_matrix.
    Xr = np.round(np.float32(4) * X) * np.float32(0.25)
    np.putmask(X, np.isclose(X, Xr), Xr)
    assert X.dtype == np.complex64
    return X.T


def polarization_matrices(outputs, inputs):
    """Compute a pair of Mueller matrices for conversion via a circular (RL)
    frame.

    Parameters
    ----------
    outputs : list
        Output polarizations
    inputs : list
        Input polarizations

    Returns
    -------
    from_circular : matrix
        Mueller matrix from circular frame to `output`
    to_circular : matrix
        Mueller matrix from `inputs` to circular frame

    Raises
    ------
    ValueError
        If `inputs` are not a full polarization basis
    """
    circular = [STOKES_RR, STOKES_RL, STOKES_LR, STOKES_LL]
    return polarization_matrix(outputs, circular), polarization_matrix(circular, inputs)


def parse_stokes(str_value):
    ans = []
    for p in str_value:
        if p not in 'IQUV':
            raise ValueError('Invalid Stokes parameter {}'.format(p))
    if not str_value:
        raise ValueError('Empty Stokes parameter list')
    for p in 'IQUV':
        cnt = str_value.count(p)
        if cnt > 1:
            raise ValueError('Stokes parameter {} listed multiple times'.format(p))
        elif cnt > 0:
            ans.append(STOKES_NAMES.index(p))
    return sorted(ans)



import pytest

from pybind11_tests import exceptions as m
import pybind11_cross_module_tests as cm


def test_std_exception(msg):
    with pytest.raises(RuntimeError) as excinfo:
        m.throw_std_exception()
    assert msg(excinfo.value) == "This exception was intentionally thrown."


def test_error_already_set(msg):
    with pytest.raises(RuntimeError) as excinfo:
        m.throw_already_set(False)
    assert msg(excinfo.value) == "Unknown internal error occurred"

    with pytest.raises(ValueError) as excinfo:
        m.throw_already_set(True)
    assert msg(excinfo.value) == "foo"


def test_cross_module_exceptions():
    with pytest.raises(RuntimeError) as excinfo:
        cm.raise_runtime_error()
    assert str(excinfo.value) == "My runtime error"

    with pytest.raises(ValueError) as excinfo:
        cm.raise_value_error()
    assert str(excinfo.value) == "My value error"

    with pytest.raises(ValueError) as excinfo:
        cm.throw_pybind_value_error()
    assert str(excinfo.value) == "pybind11 value error"

    with pytest.raises(TypeError) as excinfo:
        cm.throw_pybind_type_error()
    assert str(excinfo.value) == "pybind11 type error"

    with pytest.raises(StopIteration) as excinfo:
        cm.throw_stop_iteration()


def test_python_call_in_catch():
    d = {}
    assert m.python_call_in_destructor(d) is True
    assert d["good"] is True


def test_exception_matches():
    assert m.exception_matches()
    assert m.exception_matches_base()
    assert m.modulenotfound_exception_matches_base()


def test_custom(msg):
    # Can we catch a MyException?
    with pytest.raises(m.MyException) as excinfo:
        m.throws1()
    assert msg(excinfo.value) == "this error should go to a custom type"

    # Can we translate to standard Python exceptions?
    with pytest.raises(RuntimeError) as excinfo:
        m.throws2()
    assert msg(excinfo.value) == "this error should go to a standard Python exception"

    # Can we handle unknown exceptions?
    with pytest.raises(RuntimeError) as excinfo:
        m.throws3()
    assert msg(excinfo.value) == "Caught an unknown exception!"

    # Can we delegate to another handler by rethrowing?
    with pytest.raises(m.MyException) as excinfo:
        m.throws4()
    assert msg(excinfo.value) == "this error is rethrown"

    # Can we fall-through to the default handler?
    with pytest.raises(RuntimeError) as excinfo:
        m.throws_logic_error()
    assert msg(excinfo.value) == "this error should fall through to the standard handler"

    # Can we handle a helper-declared exception?
    with pytest.raises(m.MyException5) as excinfo:
        m.throws5()
    assert msg(excinfo.value) == "this is a helper-defined translated exception"

    # Exception subclassing:
    with pytest.raises(m.MyException5) as excinfo:
        m.throws5_1()
    assert msg(excinfo.value) == "MyException5 subclass"
    assert isinstance(excinfo.value, m.MyException5_1)

    with pytest.raises(m.MyException5_1) as excinfo:
        m.throws5_1()
    assert msg(excinfo.value) == "MyException5 subclass"

    with pytest.raises(m.MyException5) as excinfo:
        try:
            m.throws5()
        except m.MyException5_1:
            raise RuntimeError("Exception error: caught child from parent")
    assert msg(excinfo.value) == "this is a helper-defined translated exception"


def test_nested_throws(capture):
    """Tests nested (e.g. C++ -> Python -> C++) exception handling"""

    def throw_myex():
        raise m.MyException("nested error")

    def throw_myex5():
        raise m.MyException5("nested error 5")

    # In the comments below, the exception is caught in the first step, thrown in the last step

    # C++ -> Python
    with capture:
        m.try_catch(m.MyException5, throw_myex5)
    assert str(capture).startswith("MyException5: nested error 5")

    # Python -> C++ -> Python
    with pytest.raises(m.MyException) as excinfo:
        m.try_catch(m.MyException5, throw_myex)
    assert str(excinfo.value) == "nested error"

    def pycatch(exctype, f, *args):
        try:
            f(*args)
        except m.MyException as e:
            print(e)

    # C++ -> Python -> C++ -> Python
    with capture:
        m.try_catch(
            m.MyException5, pycatch, m.MyException, m.try_catch, m.MyException, throw_myex5)
    assert str(capture).startswith("MyException5: nested error 5")

    # C++ -> Python -> C++
    with capture:
        m.try_catch(m.MyException, pycatch, m.MyException5, m.throws4)
    assert capture == "this error is rethrown"

    # Python -> C++ -> Python -> C++
    with pytest.raises(m.MyException5) as excinfo:
        m.try_catch(m.MyException, pycatch, m.MyException, m.throws5)
    assert str(excinfo.value) == "this is a helper-defined translated exception"



#!/usr/bin/env python3

import argparse
import uuid

import numpy as np
from katsdpsigproc import accel
import katpoint
from astropy import units
from katsdpimager import parameters, polarization, sky_model, predict
from katsdpimager.test.utils import RandomState


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--vis', type=int, default=10**6)
    parser.add_argument('--sources', type=int, default=10**4)
    args = parser.parse_args()

    image_parameters = parameters.ImageParameters(
        q_fov=1.0,
        image_oversample=None,
        frequency=0.2 * units.m,
        array=None,
        polarizations=polarization.STOKES_IQUV,
        dtype=np.float64,
        pixel_size=0.00001,
        pixels=4096)
    oversample = 8
    w_planes = 100
    grid_parameters = parameters.GridParameters(
        antialias_width=7.0,
        oversample=oversample,
        image_oversample=4,
        w_slices=10,
        w_planes=w_planes,
        max_w=5 * units.m,
        kernel_width=7)
    base_sources = [
        "dummy0, radec, 19:39:25.03, -63:42:45.7, (200.0 12000.0 -11.11 7.777 -1.231 0 0 0 1 0.1 0 0)",       # noqa: E501
        "dummy1, radec, 19:39:20.38, -63:42:09.1, (800.0 8400.0 -3.708 3.807 -0.7202 0 0 0 1 0.2 0.2 0.2)",   # noqa: E501
        "dummy2, radec, 19:39:08.29, -63:42:33.0, (800.0 43200.0 0.956 0.584 -0.1644 0 0 0 1 0.1 0 1)"        # noqa: E501
    ]
    sources = []
    for i in range(args.sources):
        sources.append(str(uuid.uuid4()) + base_sources[i % len(base_sources)])
    model = sky_model.KatpointSkyModel(katpoint.Catalogue(sources))
    phase_centre = katpoint.construct_radec_target(
        '19:39:30', '-63:42:30').astrometric_radec() * units.rad

    rs = RandomState(seed=1)
    uv = rs.random_integers(-2048, 2048, size=(args.vis, 2)).astype(np.int16)
    sub_uv = rs.random_integers(0, grid_parameters.oversample - 1,
                                size=(args.vis, 2)).astype(np.int16)
    w_plane = rs.random_integers(0, grid_parameters.w_planes - 1, size=args.vis).astype(np.int16)
    weights = rs.uniform(size=(args.vis, len(image_parameters.polarizations))).astype(np.float32)
    vis = rs.complex_normal(size=(args.vis, len(image_parameters.polarizations)))

    context = accel.create_some_context(device_filter=lambda x: x.is_cuda)
    queue = context.create_command_queue()
    allocator = accel.SVMAllocator(context)
    template = predict.PredictTemplate(context, np.float32, len(image_parameters.polarizations))
    fn = template.instantiate(queue, image_parameters, grid_parameters,
                              args.vis, len(model), allocator=allocator)
    fn.ensure_all_bound()
    fn.num_vis = args.vis
    fn.set_coordinates(uv, sub_uv, w_plane)
    fn.set_vis(vis)
    fn.set_weights(weights)
    fn.set_sky_model(model, phase_centre)
    fn.set_w(1.2)
    fn()
    fn()
    queue.finish()


if __name__ == '__main__':
    main()



"""Data loading backend for CASA Measurement Sets."""

import argparse
import logging
from contextlib import closing

import casacore.tables
import casacore.quanta
import numpy as np
from astropy import units
import astropy.time
import astropy.coordinates
import astropy.io.fits

import katsdpimager.loader_core


_logger = logging.getLogger(__name__)


# Map casacore MFrequency::Types enum (used in MEAS_FREQ_REF column) to FITS reference frames
_MEAS_FREQ_REF_MAP = {
    0: 'SOURCE',        # REST in casacore
    1: 'LSRK',
    2: 'LSRD',
    3: 'BARYCENT',
    4: 'GEOCENTR',
    5: 'TOPOCENT',
    6: 'GALACTOC',
    7: 'LOCALGRP',
    8: 'CMBDIPOL'
}


def _get(table, name, data, casacore_units=None, astropy_units=None,
         measinfo_type=None, measinfo_ref=None):
    """Apply scaling to data loaded from a column. Scaling is done manually
    rather than with :mod:`casacore.quanta`, because the latter only operates
    (slowly) on scalars while a direct scaling is vectorised.

    Because not all writers specify the units for all quantities, units are
    allowed to be missing.

    Parameters
    ----------
    table : `casacore.tables.table`
        Table from which to load
    name : str
        Column name
    data : array-like
        Raw data loaded from the column. The leading dimension must index
        rows.
    casacore_units : str, optional
        Expected unit. A column with compatible units will be converted.
        Use ``None`` for unitless quantities.
    astropy_units : :class:`astropy.units.Unit`, optional
        Astropy unit equivalent to `casacore_units`, used for the return
        value. Use ``None`` for unitless quantities (the return value will
        then be a plain numpy array).
    measinfo_type : str, optional
        If specified, requires the ``MEASINFO`` field keyword to have this
        type.
    measinfo_ref : str, optional
        If specified, requires the ``MEASINFO`` field keyword to have this
        reference.
    """
    keywords = table.getcolkeywords(name)
    quantum_units = keywords.get('QuantumUnits')
    if quantum_units is not None:
        if casacore_units is None:
            raise ValueError('Found unexpected QuantumUnits for column {}'.format(name))
        quantum_units = np.array(quantum_units)
        if len(data.shape) == 1:
            # Scalar column. Convert quantum_units from a 1D array of 1 element
            # to a numpy scalar
            if quantum_units.shape != (1,):
                raise ValueError('Column {} is scalar, but QuantumUnits has shape {}'
                                 .format(name, quantum_units.shape))
            quantum_units = quantum_units[0]
        # data[0] is included in the iteration so that broadcasting rules are
        # triggered if quantum_units has a smaller shape (e.g. one unit for both
        # receptors on a feed).
        it = np.nditer([quantum_units, data[0]], flags=['multi_index'])
        output_quantity = casacore.quanta.quantity(1, casacore_units)
        while not it.finished:
            input_quantity = casacore.quanta.quantity(1, it[0][()])
            if not input_quantity.conforms(output_quantity):
                raise ValueError('Expected {} in column {} but found {} instead'.format(
                                 casacore_units, name, input_quantity.get_unit()))
            ratio = input_quantity.get_value(casacore_units)
            if ratio != 1:
                # Special-case ratio=1 because it is the common case, and the scaling
                # is potentially expensive.
                data[np.index_exp[:] + it.multi_index] *= ratio
            it.iternext()
    if astropy_units is not None:
        data = units.Quantity(data, astropy_units, copy=False)
    measinfo = keywords.get('MEASINFO')
    if measinfo is not None:
        if (measinfo_ref is not None and measinfo['Ref'] != measinfo_ref) or \
                (measinfo_type is not None and measinfo['type'] != measinfo_type):
            raise ValueError('Unsupported MEASINFO for {}: {}'.format(name, measinfo))
    return data


def _getcol(table, name, start=0, count=-1,
            casacore_units=None, astropy_units=None,
            measinfo_type=None, measinfo_ref=None):
    """Wrap the getcol method to fetch a batch of data from a column, applying
    scaling to the expected units. See :func:`_get` for details.

    Parameters
    ----------
    table : `casacore.tables.table`
        Table from which to load
    name : str
        Column name
    start : int
        First row to load
    count : int
        Number of rows to load
    casacore_units : str, optional
        Expected unit. A column with compatible units will be converted.
        Use ``None`` for unitless quantities.
    astropy_units : :class:`astropy.units.Unit`, optional
        Astropy unit equivalent to `casacore_units`, used for the return
        value. Use ``None`` for unitless quantities (the return value will
        then be a plain numpy array).
    measinfo_type : str, optional
        If specified, requires the ``MEASINFO`` field keyword to have this
        type.
    measinfo_ref : str, optional
        If specified, requires the ``MEASINFO`` field keyword to have this
        reference.
    """
    data = table.getcol(name, start, count)
    return _get(table, name, data, casacore_units, astropy_units, measinfo_type, measinfo_ref)


def _getcolslice(table, name, blc, trc, inc=[], start=0, count=-1,
                 casacore_units=None, astropy_units=None,
                 measinfo_type=None, measinfo_ref=None):
    """Wrap the getcol method to fetch a batch of data from a column with
    slicing, applying scaling to the expected units. See :func:`_get` for
    details.

    Parameters
    ----------
    table : `casacore.tables.table`
        Table from which to load
    name : str
        Column name
    blc : list
        Bottom left corner for array slice
    trc : list
        Top right corner for array slice (inclusive)
    ind : list
        Strides
    start : int
        First row to load
    count : int
        Number of rows to load
    casacore_units : str, optional
        Expected unit. A column with compatible units will be converted.
        Use ``None`` for unitless quantities.
    astropy_units : :class:`astropy.units.Unit`, optional
        Astropy unit equivalent to `casacore_units`, used for the return
        value. Use ``None`` for unitless quantities (the return value will
        then be a plain numpy array).
    measinfo_type : str, optional
        If specified, requires the ``MEASINFO`` field keyword to have this
        type.
    measinfo_ref : str, optional
        If specified, requires the ``MEASINFO`` field keyword to have this
        reference.
    """
    data = table.getcolslice(name, blc, trc, inc, start, count)
    return _get(table, name, data, casacore_units, astropy_units, measinfo_type, measinfo_ref)


def _getcolchannels(table, name, start_channel, stop_channel, start=0, count=-1,
                    casacore_units=None, astropy_units=None,
                    measinfo_type=None, measinfo_ref=None):
    """Convenience wrapper around :func:`_getcolslice` for columns where the
    shape is channels by polarizations, and a range of channels is desired.
    """
    return _getcolslice(table, name, [start_channel, -1], [stop_channel - 1, -1], [],
                        start, count, casacore_units, astropy_units,
                        measinfo_type, measinfo_ref)


def _getcell(table, name, row,
             casacore_units=None, astropy_units=None,
             measinfo_type=None, measinfo_ref=None):
    """Like :meth:`_getcol`, but for a single cell"""
    data = table.getcell(name, row)
    # Temporary add a leading dimension, required by _get
    data = data[np.newaxis, :]
    data = _getcol(table, name, row, 1, casacore_units, astropy_units,
                   measinfo_type, measinfo_ref)
    out = _get(table, name, data, casacore_units, astropy_units, measinfo_type, measinfo_ref)
    return out[0]


def _fix_cache_size(table, column):
    """Workaround for https://github.com/casacore/casacore/issues/581.
    We set the column cache size to be large enough for two tiles worth of
    rows.
    """
    cache_size = 0
    try:
        for item in table.getdminfo(column)['SPEC']['HYPERCUBES'].values():
            cube_shape = tuple(item['CubeShape'])
            tile_shape = tuple(item['TileShape'])
            cache_shape = cube_shape[:2] + tile_shape[2:]
            cache_size = max(cache_size, 2 * int(np.product(cache_shape)))
    except (TypeError, KeyError, IndexError, ValueError):
        pass
    if cache_size == 0:
        # No tile size info found - pick some default
        _logger.warning('Could not get tile info for column %s', column)
        cache_size = 1024 * 1024
    _logger.debug('Using cache size %d for column %s', cache_size, column)
    table.setmaxcachesize(column, cache_size)


class LoaderMS(katsdpimager.loader_core.LoaderBase):
    def __init__(self, filename, options):
        super().__init__(filename, options)
        parser = argparse.ArgumentParser(
            prog='Measurement set options',
            usage='Measurement set options: [-i data=COLUMN] [-i field=FIELD] ...')
        parser.add_argument('--data', type=str, metavar='COLUMN', default='DATA',
                            help='Column containing visibilities to image [%(default)s]')
        parser.add_argument('--data-desc', type=int, default=0,
                            help='Data description ID to image [%(default)s]')
        parser.add_argument('--field', type=int, default=0,
                            help='Field to image [%(default)s]')
        parser.add_argument('--pol-frame', choices=['sky', 'feed'], default='sky',
                            help='Reference frame for polarization [%(default)s]')
        parser.add_argument('--uvw', choices=['casa', 'strict'], default='casa',
                            help='UVW sign convention [%(default)s]')
        args = parser.parse_args(options)
        self._strict_uvw = (args.uvw == 'strict')
        self._main = casacore.tables.table(filename, ack=False)
        _fix_cache_size(self._main, 'FLAG')
        self._antenna = casacore.tables.table(self._main.getkeyword('ANTENNA'), ack=False)
        self._data_description = casacore.tables.table(self._main.getkeyword('DATA_DESCRIPTION'),
                                                       ack=False)
        self._field = casacore.tables.table(self._main.getkeyword('FIELD'), ack=False)
        self._spectral_window = casacore.tables.table(self._main.getkeyword('SPECTRAL_WINDOW'),
                                                      ack=False)
        self._polarization = casacore.tables.table(self._main.getkeyword('POLARIZATION'), ack=False)
        self._feed = casacore.tables.table(self._main.getkeyword('FEED'), ack=False)
        self._data_col = args.data
        self._field_id = args.field
        self._data_desc_id = args.data_desc
        if self._data_col not in self._main.colnames():
            raise ValueError('{} has no column named {}'.format(
                filename, self._data_col))
        if args.field < 0 or args.field >= self._field.nrows():
            raise ValueError('Field {} is out of range'.format(args.field))
        if args.data_desc < 0 or args.data_desc >= self._data_description.nrows():
            raise ValueError('Data description {} is out of range'.format(args.data_desc))
        self._polarization_id = self._data_description.getcell(
            'POLARIZATION_ID', args.data_desc)
        self._spectral_window_id = self._data_description.getcell(
            'SPECTRAL_WINDOW_ID', args.data_desc)
        # Detect whether we have a usable WEIGHT_SPECTRUM column. Measurement
        # sets in the wild sometimes have the column but with 0x0 shape.
        try:
            self._main.getcellslice('WEIGHT_SPECTRUM', 0, [0, 0], [0, 0])
        except RuntimeError:
            self._has_weight_spectrum = False
        else:
            self._has_weight_spectrum = True

        self._feed_angle_correction = args.pol_frame == 'feed'
        if self._feed_angle_correction:
            # Load per-antenna feed angles. For now, only a constant value per
            # antenna is supported, rather than per feed or per receptor within a
            # feed. This avoids the need for more per-visibility indexing.
            antenna_id = _getcol(self._feed, 'ANTENNA_ID')
            receptor_angle = _getcol(self._feed, 'RECEPTOR_ANGLE', 0, -1, 'rad', units.rad)
            antenna_angle = [None] * (np.max(antenna_id) + 1)
            for i in range(len(antenna_id)):
                for angle in receptor_angle[i]:
                    if (antenna_angle[antenna_id[i]] is not None
                            and not np.allclose(antenna_angle[antenna_id[i]],
                                                angle, atol=1e-8 * units.rad)):
                        raise ValueError('Multiple feed angles for one antenna is not supported')
                    antenna_angle[antenna_id[i]] = angle
            self._antenna_angle = units.Quantity(antenna_angle)
        else:
            self._antenna_angle = None
        self._average_time = None     # Will be set by data_iter
        self._observation_ids = set()

    @classmethod
    def match(cls, filename):
        return filename.lower().endswith('.ms')

    def antenna_diameters(self):
        return _getcol(self._antenna, 'DISH_DIAMETER', 0, -1, 'm', units.m)

    def antenna_positions(self):
        return _getcol(self._antenna, 'POSITION', 0, -1, 'm', units.m, 'position', 'ITRF')

    def phase_centre(self):
        value = _getcell(self._field, 'PHASE_DIR', self._field_id,
                         'rad', units.rad, 'direction', 'J2000')
        if tuple(value.shape) != (1, 2):
            raise ValueError('Unsupported shape for PHASE_DIR: {}'.format(value.shape))
        return value[0, :]

    def num_channels(self):
        return len(_getcell(self._spectral_window, 'CHAN_FREQ', self._spectral_window_id,
                            'Hz', units.Hz))

    def frequency(self, channel):
        return _getcell(self._spectral_window, 'CHAN_FREQ', self._spectral_window_id,
                        'Hz', units.Hz)[channel]

    def polarizations(self):
        return _getcell(self._polarization, 'CORR_TYPE', self._polarization_id)

    def has_feed_angles(self):
        return self._feed_angle_correction

    def extra_fits_headers(self):
        headers = astropy.io.fits.Header()
        obsgeo = np.mean(self.antenna_positions().to_value(units.m), axis=0)
        obsgeo_comment = 'Average of antenna positions'
        headers['OBSGEO-X'] = (obsgeo[0], obsgeo_comment)
        headers['OBSGEO-Y'] = (obsgeo[1], obsgeo_comment)
        headers['OBSGEO-Z'] = (obsgeo[2], obsgeo_comment)

        if self._average_time is not None:
            headers['DATE-AVG'] = self._average_time.utc.isot

        meas_freq_ref = self._spectral_window.getcell('MEAS_FREQ_REF', self._spectral_window_id)
        if meas_freq_ref in _MEAS_FREQ_REF_MAP:
            headers['SPECSYS'] = _MEAS_FREQ_REF_MAP[meas_freq_ref]

        if len(self._observation_ids) == 1:
            row = list(self._observation_ids)[0]
            table = casacore.tables.table(self._main.getkeyword('OBSERVATION'), ack=False)
            with closing(table):
                time_range = _getcell(table, 'TIME_RANGE', row,
                                      's', None, 'epoch', 'UTC')
                observer = table.getcell('OBSERVER', row)
                telescope = table.getcell('TELESCOPE_NAME', row)
            start_time = astropy.time.Time(time_range[0] / 86400.0, format='mjd', scale='utc')
            headers['DATE-OBS'] = start_time.utc.isot
            headers['TELESCOP'] = telescope
            headers['OBSERVER'] = observer
        elif len(self._observation_ids) > 1:
            _logger.warning('Multiple OBSERVATION_IDs; will not add FITS headers for observation')

        return headers

    def data_iter(self, start_channel, stop_channel, max_chunk_vis=None):
        if max_chunk_vis is None:
            max_chunk_vis = self._main.nrows()
        num_channels = stop_channel - start_channel
        max_chunk_rows = self._main.nrows()
        if max_chunk_vis is not None:
            max_chunk_rows = max(1, max_chunk_vis // num_channels)
        # PHASE_DIR is used as an approximation to the antenna pointing
        # direction, for the purposes of parallactic angle correction. This
        # probably doesn't generalise well beyond dishes with single-pixel
        # feeds.
        pointing = self.phase_centre()
        pointing = astropy.coordinates.SkyCoord(ra=pointing[0], dec=pointing[1], frame='fk5')
        pos = self.antenna_positions()
        pos = astropy.coordinates.EarthLocation.from_geocentric(pos[:, 0], pos[:, 1], pos[:, 2])
        # Construct a point that is displaced from the pointing by a small
        # quantity northwards. It's necessary to use a small finite difference
        # rather than the pole itself, because the transformation to AzEl is
        # not rigid (does not preserve great circles).
        pole = pointing.directional_offset_by(0 * units.rad, 1e-5 * units.rad)
        time_sum = 0.0
        time_count = 0
        for start in range(0, self._main.nrows(), max_chunk_rows):
            end = min(self._main.nrows(), start + max_chunk_rows)
            nrows = end - start
            flag_row = _getcol(self._main, 'FLAG_ROW', start, nrows)
            field_id = _getcol(self._main, 'FIELD_ID', start, nrows)
            data_desc_id = _getcol(self._main, 'DATA_DESC_ID', start, nrows)
            antenna1 = _getcol(self._main, 'ANTENNA1', start, nrows)
            antenna2 = _getcol(self._main, 'ANTENNA2', start, nrows)
            observation_id = _getcol(self._main, 'OBSERVATION_ID', start, nrows)
            valid = np.logical_not(flag_row)
            valid = np.logical_and(valid, field_id == self._field_id)
            valid = np.logical_and(valid, data_desc_id == self._data_desc_id)
            valid = np.logical_and(valid, antenna1 != antenna2)
            self._observation_ids.update(observation_id)
            antenna1 = antenna1[valid]
            antenna2 = antenna2[valid]
            data = _getcolchannels(self._main, self._data_col, start_channel, stop_channel,
                                   start, nrows, 'Jy', units.Jy)[valid, ...]
            feed_angle1 = None
            feed_angle2 = None
            time_full = _getcol(self._main, 'TIME_CENTROID', start, nrows, 's', None,
                                'epoch', 'UTC')[valid, ...]
            # Each time will be repeated per baseline, but we do not need to repeat all the
            # calculations for each time. Extract just the unique times.
            time, inverse = np.unique(time_full, return_inverse=True)
            time_sum += np.sum(time)
            time_count += len(time)
            # Update average time as we go, in case the caller doesn't exhaust the generator
            average_time = time_sum / time_count
            self._average_time = astropy.time.Time(
                average_time / 86400.0, format='mjd', scale='utc')
            if self._feed_angle_correction:
                # Convert time from MJD seconds to MJD. We do this here rather
                # than by passing 'd' to _getcol, because not all measurement sets
                # specify the units and we want to assume seconds if not specified.
                time = astropy.time.Time(time / 86400.0, format='mjd', scale='utc')
                # Convert to CIRS, which is closer to AltAz in astropy's
                # conversion graph. This avoids duplicating precession-nutation
                # calculations for every antenna.
                cirs_frame = astropy.coordinates.CIRS(obstime=time)
                pole_cirs = pole.transform_to(cirs_frame)
                pointing_cirs = pointing.transform_to(cirs_frame)
                feed_angle = units.Quantity(
                    np.empty((len(time), len(pos))), unit=units.rad, copy=False)
                for i in range(len(pos)):
                    altaz_frame = astropy.coordinates.AltAz(location=pos[i], obstime=time)
                    pole_altaz = pole_cirs.transform_to(altaz_frame)
                    pointing_altaz = pointing_cirs.transform_to(altaz_frame)
                    pa = pointing_altaz.position_angle(pole_altaz)
                    feed_angle[:, i] = pa + self._antenna_angle[i]
                feed_angle1 = feed_angle[inverse, antenna1].astype(np.float32)
                feed_angle2 = feed_angle[inverse, antenna2].astype(np.float32)
            uvw = _getcol(self._main, 'UVW', start, nrows, 'm', units.m, 'uvw')[valid, ...]
            if not self._strict_uvw:
                uvw = -uvw
            if self._has_weight_spectrum:
                weight = _getcolchannels(self._main, 'WEIGHT_SPECTRUM', start_channel, stop_channel,
                                         start, nrows)[valid, ...]
            else:
                weight = _getcol(self._main, 'WEIGHT', start, nrows)[valid, ...]
                # Fake a channel axis
                weight = weight[:, np.newaxis, :]
                weight = np.broadcast_to(weight, data.shape)
            flag = _getcolchannels(self._main, 'FLAG', start_channel, stop_channel,
                                   start, nrows)[valid, ...]
            weight = weight * np.logical_not(flag)
            baseline = (antenna1 * self._antenna.nrows() + antenna2)
            ret = dict(uvw=uvw,
                       weights=np.swapaxes(weight, 0, 1),
                       baselines=baseline,
                       vis=np.swapaxes(data, 0, 1),
                       progress=end,
                       total=self._main.nrows())
            if self._feed_angle_correction:
                ret['feed_angle1'] = feed_angle1
                ret['feed_angle2'] = feed_angle2
            yield ret

    @property
    def raw_data(self):
        return self._main

    def close(self):
        self._main.close()
        self._antenna.close()
        self._data_description.close()
        self._field.close()
        self._spectral_window.close()
        self._polarization.close()
        self._feed.close()



#!/usr/bin/env python3

import argparse

import katsdpimager.render as render


def main():
    parser = argparse.ArgumentParser(
        description='Create video from FITS files written by katsdpimager.'
    )
    parser.add_argument('--width', type=int, default=1024,
                        help='Video width [%(default)s]')
    parser.add_argument('--height', type=int, default=768,
                        help='Video height [%(default)s]')
    parser.add_argument('--fps', type=float, default=5.0,
                        help='Frames per second [%(default)s]')
    parser.add_argument('output', help='Output video file')
    parser.add_argument('input', nargs='+', help='Input FITS files')
    args = parser.parse_args()

    files = [(None, arg) for arg in args.input]
    render.write_movie(files, args.output, width=args.width, height=args.height, fps=args.fps)


if __name__ == '__main__':
    main()
